/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/


/*==================================================================
 *                       6.1 BOOLEANS
 *==================================================================
 * Standard scheme R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 *  not                 OBJ -> BOOLEAN
 *  boolean?            OBJ -> BOOLEAN
 *
 *==================================================================
 */

#include "swit.h"

SWiT swit_true;
SWiT swit_false;

static SWiT not_prim (SWiT arg);
static SWiT boolean_p_prim (SWiT arg);

static void
print_boolean (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  if (obj == swit_true) {
    if (tcl) {
      str2buff("1");
    } else {
      str2buff("#t");
    }
  } else {
    if (tcl) {
      str2buff("0");
    } else {
      str2buff("#f");
    }
  }
}

static SWiT
make_boolean (int val)
{
  SWiT obj = swit_alloc_object();
  SWIT_TYPE(obj) = BOOLEAN;
  obj->data = (int *)val;
  return (obj);
}

static SWiT
not_prim (SWiT arg)
{
  return ((arg == swit_false) ?
	  swit_true : swit_false);
}

static SWiT
boolean_p_prim (SWiT arg)
{
  return (((arg == swit_false) || (arg == swit_true)) ?
	  swit_true : swit_false);
}

/*==================================================================
 *                       INITIALISATION
 *==================================================================
 */

void
swit_init_boolean (void)
{
  print_table[BOOLEAN>>1] = print_boolean;
  swit_true = make_boolean(1);
  swit_false = make_boolean(0);
  swit_make_prim1 ("not", not_prim, ANY);
  swit_make_prim1 ("boolean?", boolean_p_prim, ANY);
}
