/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "swit.h"
#include "code.h"

static SWiT inst[100];
/* static SWiT decode (SWiT_Expr e); */

SWiT_Expr
swit_make_quote (SWiT val)
{
  if (val && SWIT_PAIRP(val)) {
    SWiT_Expr expr = swit_alloc_expr();
    SWIT_INST(expr) = QUOTE_CODE;
    SWIT_QUOTE(expr) = val;
    return (expr);
  } else {
    return (SWiT_Expr)val;
  }
}

/*
static SWiT
decode_quote (SWiT_Expr expr)
{
  return swit_make_pair(inst[QUOTE_CODE], SWIT_QUOTE(expr));
}
*/
static SWiT_Expr
make_step_code (SWiT_Expr expr0)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWIT_INST(expr) = STEP_CODE;
  SWIT_CODE(expr) = expr0;
  return (expr);
}

/* TODO ...
static SWiT
decode_step (SWiT_Expr expr)
{
  return swit_make_pair(inst[STEP_CODE], decode(SWIT_CODE(expr)));
}
*/

static SWiT_Expr
make_symbol_code (SWiT symbol)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWIT_INST(expr) = SYMBOL_CODE;
  SWIT_SYMBOL(expr) = SWIT_SYM(symbol);
  return expr;
}

/*
static SWiT
decode_symbol (SWiT_Expr expr)
{
  return swit_make_pair(inst[SYMBOL_CODE], SWIT_SYM(expr));
}
*/

static SWiT_Expr
make_lvar_code (SWiT sdb, SWiT sidx)
{
  unsigned int db, idx;
  SWiT_Expr expr;
  db = SWIT_FIXNUM(sdb);
  idx = SWIT_FIXNUM(sidx);
  expr = swit_alloc_expr();
  SWIT_INST (expr) = LVAR_CODE;
  SWIT_VAR(expr) = db + (idx << 16);
  return (expr);
}

static SWiT_Expr
make_lambda_code (SWiT formals, SWiT_Expr body, SWiT pcode)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Lambda lbd = SWIT_ALLOC(Lambda);
  SWiT arity = SWIT_CAR(formals);
  SWiT params = SWIT_CDR(formals);
  int len;
  LAMBDA_NAMES(lbd) = params; /* should be checked */
  LAMBDA_PCODE(lbd) = pcode;
  len = SWIT_FIXNUM(arity);
  LAMBDA_NPARAMS(lbd) =
    ((len < 0) || ((len == 0) && (params != swit_null_list))) ?
    -len + (1 << 16) : len;
  LAMBDA_BODY(lbd) = body;
  SWIT_INST(expr) = LAMBDA_CODE;
  SWIT_LAMBDA(expr) = lbd;
  return (expr);
}

/*exported to switControl.c (for swit_apply) */
SWiT_Expr
make_call_code (SWiT_Expr fun, int len, SWiT_Expr *args)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Call call_code = SWIT_ALLOC(Call);

  CALL_PROC (call_code) = fun;
  CALL_NARGS(call_code) = len;
  CALL_ARGS (call_code) = args;
  if (len == 0) {
    SWIT_INST (expr) = THUNK_CODE;
  } else {
    SWIT_INST (expr) = CALL_CODE;
  }
  SWIT_CALL(expr) = call_code;
  return (expr);
}

static SWiT_Expr
make_call0_code (SWiT_Sym_prim0 prim)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWIT_INST (expr) = CALL0_CODE;
  SWIT_CALL0 (expr) = prim;
  return (expr);
}

static SWiT_Expr
make_call0n_code (SWiT_Sym_prim0n prim, int len, SWiT_Expr *args)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Call0n cll = SWIT_ALLOC(Call0n);

  CALL_PRIM (cll) = prim;
  CALL_NARGS (cll) = len;
  CALL_ARGS (cll) = args;
  SWIT_INST (expr) = CALL0N_CODE;
  SWIT_CALL0n (expr) = cll;
  return (expr);
}

static SWiT_Expr
make_call1_code (SWiT_Sym_prim1 prim, SWiT_Expr arg)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Call1 cll = SWIT_ALLOC(Call1);

  CALL_PRIM (cll) = prim;
  CALL_ARG (cll) = arg;
  SWIT_INST (expr) = CALL1_CODE;
  SWIT_CALL1 (expr) = cll;
  return (expr);
}

static SWiT_Expr
make_call1n_code (SWiT_Sym_prim1n prim, SWiT_Expr arg, int argc, SWiT_Expr *argv)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Call1n cll = SWIT_ALLOC(Call1n);

  CALL_PRIM (cll) = prim;
  CALL_ARG (cll) = arg;
  CALL_NARGS (cll) = argc;
  CALL_ARGS (cll) = argv;
  SWIT_INST (expr) = CALL1N_CODE;
  SWIT_CALL1n(expr) = cll;
  return (expr);
}

static SWiT_Expr
make_call2_code (SWiT_Sym_prim2 prim, SWiT_Expr arg1, SWiT_Expr arg2)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Call2 cll = SWIT_ALLOC(Call2);

  CALL_PRIM (cll) = prim;
  CALL_ARG1 (cll) = arg1;
  CALL_ARG2 (cll) = arg2;
  SWIT_INST (expr) = CALL2_CODE;
  SWIT_CALL2(expr) = cll;
  return (expr);
}

static SWiT_Expr
make_call2n_code (SWiT_Sym_prim2n prim, SWiT_Expr arg1, SWiT_Expr arg2, int argc, SWiT_Expr *argv)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Call2n cll = SWIT_ALLOC(Call2n);

  CALL_PRIM (cll) = prim;
  CALL_ARG1 (cll) = arg1;
  CALL_ARG2 (cll) = arg2;
  CALL_NARGS (cll) = argc;
  CALL_ARGS (cll) = argv;
  SWIT_INST (expr) = CALL2N_CODE;
  SWIT_CALL2n (expr) = cll;
  return (expr);
}

static SWiT_Expr
make_call3_code (SWiT_Sym_prim3 prim, SWiT_Expr *args)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Call3 cll = SWIT_ALLOC(Call3);

  CALL_PRIM (cll) = prim;
  CALL_ARG1 (cll) = args[0];
  CALL_ARG2 (cll) = args[1];
  CALL_ARG3 (cll) = args[2];
  SWIT_INST (expr) = CALL3_CODE;
  SWIT_CALL3 (expr) = cll;
  return (expr);
}

static SWiT_Expr
make_pcall_code (SWiT fun, int len, SWiT_Expr *args)
{
  switch SWIT_TYPE(fun)
    {
    case PRIM0:
      return (make_call0_code(SWIT_PRIM0(fun)));
    case PRIM0N:
      return (make_call0n_code(SWIT_PRIM0n(fun), len, args));
    case PRIM01:
      if (len == 0) {
	return (make_call1_code(SWIT_PRIM1(fun), swit_make_quote(PRIM_VAL(SWIT_PRIM1(fun)))));
      /*      return (make_call1_code(SWIT_PRIM1(fun), PRIM_VAL(SWIT_PRIM1(fun))));*/
      } /* else */
    case PRIM1:
      return (make_call1_code(SWIT_PRIM1(fun), *args));
    case PRIM1N:
      return (make_call1n_code(SWIT_PRIM1n(fun), *args, len-1, args+1));
    case PRIM12:
      if (len == 1) {
	return (make_call2_code(SWIT_PRIM2(fun), *args, swit_make_quote(PRIM_VAL(SWIT_PRIM2(fun)))));
	/*	return (make_call2_code(SWIT_PRIM2(fun), *args, PRIM_VAL(SWIT_PRIM2(fun)))); */
      } /* else */
    case PRIM2:
      return (make_call2_code(SWIT_PRIM2(fun), args[0], args[1]));
    case PRIM2N:
      return (make_call2n_code(SWIT_PRIM2n(fun), args[0], args[1], len-2, args+2));
    case PRIM3:
      return (make_call3_code(SWIT_PRIM3(fun), args));
    }
  return NULL; /* to avoid a compiler warning */
}


static SWiT_Expr
make_if_code (SWiT_Expr test, SWiT_Expr pthen, SWiT_Expr pelse)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_If if_code = SWIT_ALLOC(If);
  IF_TEST(if_code) = test;
  IF_THEN(if_code) = pthen;
  IF_ELSE(if_code) = pelse;
  SWIT_INST(expr) = IF_CODE;
  SWIT_IF(expr) = if_code;
  return expr;
}

static SWiT_Expr
make_protect_code (SWiT_Expr prt)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWIT_INST(expr) = PROTECT_CODE;
  SWIT_CODE(expr) = prt;
  return (expr);
}


static SWiT_Expr
make_try_code (SWiT_Expr ex, SWiT_Expr *handlers, SWiT exn_lst)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Try try_code = SWIT_ALLOC(Try);
  TRY_EXPR(try_code) = ex;
  TRY_HDL(try_code) = handlers+1;
  TRY_PROTECT(try_code) = make_protect_code(*handlers);
  TRY_EXN(try_code) = exn_lst;
  SWIT_INST(expr) = TRY_CODE;
  SWIT_TRY(expr) = try_code;
  return expr;
}


static SWiT_Expr
make_gset (SWiT_Expr sym, SWiT_Expr arg)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Gset gset_code = SWIT_ALLOC(Gset);
  GSET_SYM (gset_code) = SWIT_SYMBOL(sym);
  GSET_EXPR (gset_code) = arg;
  SWIT_INST (expr) = GSET_CODE;
  SWIT_GSET(expr) = gset_code;
  return (expr);
}

static SWiT_Expr
make_lset (SWiT_Expr lvar, SWiT_Expr arg)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Lset lset_code = SWIT_ALLOC(Lset);
  LSET_VAR (lset_code) = SWIT_VAR(lvar);
  LSET_EXPR(lset_code) = arg;
  SWIT_INST (expr) = LSET_CODE;
  SWIT_LSET(expr) = lset_code;
  return (expr);
}

static SWiT_Expr
make_set_code (SWiT_Expr var, SWiT_Expr expr)
{
  int etype = SWIT_INST(var);
  if (etype == SYMBOL_CODE) {
    return make_gset (var, expr);
  } else { /* LVAR_CODE */
    return make_lset (var, expr);
  }
}

static SWiT_Expr
make_or_code (int nexprs, SWiT_Expr *exprs)
{
  SWiT_Expr or_code = swit_alloc_expr();
  SWiT_Begin code = SWIT_ALLOC(Begin);
  BEG_NFORMS (code) = nexprs-1;
  BEG_FORMS (code) = exprs;
  SWIT_BEGIN(or_code) = code;
  SWIT_INST(or_code) = OR_CODE;
  return(or_code);
}

static SWiT_Expr
make_and_code (int nexprs, SWiT_Expr *exprs)
{
  SWiT_Expr and_code = swit_alloc_expr();
  SWiT_Begin code = SWIT_ALLOC(Begin);
  BEG_NFORMS (code) = nexprs-1;
  BEG_FORMS (code) = exprs;
  SWIT_BEGIN(and_code) = code;
  SWIT_INST(and_code) = AND_CODE;
  return(and_code);
}

static SWiT_Expr
make_arrow_code(SWiT_Expr exp)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWIT_INST(expr) = ARROW_CODE;
  SWIT_CODE(expr) = exp;
  return(expr);
}

static SWiT_Expr
make_cond_code (int nclauses, SWiT_Expr *tests, SWiT_Expr *seqs)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Cond cond_code = SWIT_ALLOC(Cond);
  COND_NCLAUSES(cond_code) = nclauses;
  COND_TESTS(cond_code) = tests;
  COND_SEQS(cond_code) = seqs;
  SWIT_INST(expr) = COND_CODE;
  SWIT_COND(expr) = cond_code;
  return(expr);
}

static SWiT_Expr
make_case_code (SWiT_Expr key, int nclauses, SWiT *tests, SWiT_Expr *seqs)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Case case_code = SWIT_ALLOC(Case);
  SWIT_INST(expr) = CASE_CODE;
  CASE_KEY(case_code) = key;
  CASE_NCLAUSES(case_code) = nclauses;
  CASE_TESTS(case_code) = tests;
  CASE_SEQS(case_code) = seqs;
  SWIT_CASE(expr) = case_code;
  return(expr);
}

static SWiT_Expr
make_let_code (SWiT_Expr lbd, SWiT_Expr *args)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Let let_code = SWIT_ALLOC(Let);
  LET_LBD (let_code) = SWIT_LAMBDA(lbd);
  LET_ARGS (let_code) = args;
  SWIT_INST (expr) = LET_CODE;
  SWIT_LET(expr) = let_code;
  return (expr);
}

static SWiT_Expr
make_letstar_code (SWiT_Expr lbd, SWiT_Expr *args)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Let letstar_code = SWIT_ALLOC(Let);
  SWIT_INST (expr) = LETSTAR_CODE;
  LET_LBD (letstar_code) = SWIT_LAMBDA(lbd);
  LET_ARGS (letstar_code) = args;
  SWIT_LET(expr) = letstar_code;
  return (expr);
}

static SWiT_Expr
make_letrec_code (SWiT_Expr lbd, SWiT_Expr *args)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Let letrec_code = SWIT_ALLOC(Let);
  LET_LBD (letrec_code) = SWIT_LAMBDA(lbd);
  LET_ARGS (letrec_code) = args;
  SWIT_INST (expr) = LETREC_CODE;
  SWIT_LET(expr) = letrec_code;
  return (expr);
}

static SWiT_Expr
make_begin_code (int nforms, SWiT_Expr *forms)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Begin code = SWIT_ALLOC(Begin);

  BEG_NFORMS (code) = nforms-1;
  BEG_FORMS (code) = forms;
  SWIT_INST (expr) = BEGIN_CODE;
  SWIT_BEGIN(expr) = code;
  return (expr);
}


static SWiT_Expr
make_delay_code (SWiT_Expr arg)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWIT_INST(expr) = DELAY_CODE;
  SWIT_CODE(expr) = arg;
  return (expr);
}


#define UNSET_SYMBOL_FLAG(sym, flag) (((unsigned int)(sym)->next) &= ~flag)
static SWiT_Expr
make_define_code (SWiT symbol, SWiT_Expr arg)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Define def_code =  SWIT_ALLOC(Define);
  SWiT_Symbol sym = SWIT_SYM(symbol);
  if SYMBOL_MACROP(sym) {
    UNSET_SYMBOL_FLAG(sym, SYMBOL_MACRO_MASK);
  }
  SWIT_INST (expr) = DEFINE_CODE;
  DEFINE_SYM(def_code) =  symbol;
  DEFINE_EXPR(def_code) = arg;
  SWIT_DEFINE (expr) = def_code;
  return (expr);
}

static SWiT_Expr
make_defmacro_code (SWiT symbol, SWiT_Expr arg)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWiT_Define def_code =  SWIT_ALLOC(Define);
  SET_SYMBOL_FLAG((SWiT_Symbol)symbol->data, SYMBOL_MACRO_MASK);
  SWIT_INST (expr) = DEFMACRO_CODE;
  DEFINE_SYM(def_code) =  symbol;
  DEFINE_EXPR(def_code) = arg;
  SWIT_DEFINE (expr) = def_code;
  return (expr);
}

SWiT_Expr compile_lcode (SWiT vect);

/* exported to switType.c */
SWiT swit_two;
SWiT swit_three;

static SWiT_Expr *
compile_args (SWiT args)
{
  int len = VECTOR_LENGTH(args);
  SWiT *els = VECTOR_ELS(args);
  SWiT_Expr *code = swit_alloc_exprs(len);
  int i;
  for (i = 0; i < len; i++) {
    code[i] = compile_lcode(els[i]);
  }
  return code;
}

static SWiT_Expr *
compile_rargs (SWiT args)
{
  int len = VECTOR_LENGTH(args);
  SWiT *els = VECTOR_ELS(args);
  SWiT_Expr *code = swit_alloc_exprs(len);
  int i;
  code[0] = compile_lcode(els[0]);
  for (i = 1; i < len; i++) {
    code[i] = compile_lcode(els[len - i]);
  }
  return code;
}

#define FST(code) SWIT_CAR(code)
#define SND(code) SWIT_CDR(code)
#define TRD(code) SWIT_CDR(SWIT_CDR(code))

SWiT_Expr
compile_lcode (SWiT c)
{
  if SWIT_PAIRP(c) {
    SWiT car = SWIT_CAR(c);
    SWiT code = SWIT_CDR(c);
    int icode;
    icode = SWIT_FIXNUM(car);
    switch (icode)
      {
      case QUOTE_CODE:
	return (swit_make_quote(code));
      case SCALL_CODE:
	{
	  SWiT args = SND(code);
	  return (make_call_code(make_symbol_code(FST(code)),
				 VECTOR_LENGTH(args),
				 compile_args(args)));
	}
      case PCALL_CODE:
	{
	  SWiT args = SND(code);
	  return (make_pcall_code(FST(code),
				  VECTOR_LENGTH(args),
				  compile_args(args)));
	}
      case ECALL_CODE:
      case LET_CODE:
	{
	  SWiT args = SND(code);
	  return (make_call_code(compile_lcode(FST(code)),
				 VECTOR_LENGTH(args),
				 compile_args(args)));
	}
      case SYMBOL_CODE:
	return (make_symbol_code(code));
      case LVAR_CODE:
	return (make_lvar_code(FST(code), SND(code)));
      case LAMBDA_CODE:
	return (make_lambda_code(FST(code),
				 compile_lcode(SND(code)),
				 c));
      case IF_CODE:
	{
	  SWiT_Expr te, th, el;
	  te = compile_lcode(FST(code));
	  th = compile_lcode(FST(SND(code)));
	  el = compile_lcode(TRD(code));
	  return (make_if_code(te, th, el));
	}
      case TRY_CODE:
	{
	  SWiT_Expr expr, *handlers;
	  expr = compile_lcode(FST(code));
	  handlers = compile_args(FST(SND(code)));
	  return make_try_code(expr, handlers, TRD(code));
	}
      case AND_CODE:
	return (make_and_code(VECTOR_LENGTH(code),
			      compile_rargs(code)));
      case OR_CODE:
	return (make_or_code(VECTOR_LENGTH(code),
			     compile_rargs(code)));
      case BEGIN_CODE:
	return (make_begin_code(VECTOR_LENGTH(code),
				/* liste renverse aprs le premier */
				compile_rargs(code)));
      case ARROW_CODE:
	return (make_arrow_code(compile_lcode(code)));
      case COND_CODE:
	{
	  SWiT args = FST(code);
	  return (make_cond_code(VECTOR_LENGTH(args),
				 compile_args(args),
				 compile_args(SND(code))));
	}
      case CASE_CODE:
	{
	  SWiT_Expr key;
	  SWiT args;
	  key = compile_lcode(FST(code));
	  args = FST(SND(code));
	  return (make_case_code(key,
				 VECTOR_LENGTH(args),
				 VECTOR_ELS(args),
				 compile_args(TRD(code))));
	}
      case SET_CODE:
	return (make_set_code(compile_lcode(FST(code)),
			      compile_lcode(SND(code))));
	/*      case LET_CODE:
		return (make_let_code(compile_lcode(FST(code)),
		compile_args(SND(code))));*/
      case LETSTAR_CODE:
	return (make_letstar_code(compile_lcode(FST(code)),
				  compile_args(SND(code))));
      case LETREC_CODE:
	return (make_letrec_code(compile_lcode(FST(code)),
				 compile_args(SND(code))));
      case DELAY_CODE:
	return (make_delay_code(compile_lcode(code)));
      case DEFMACRO_CODE:
	return (make_defmacro_code(FST(code),
				   compile_lcode(SND(code))));
      case DEFINE_CODE:
	return (make_define_code(FST(code),
				 compile_lcode(SND(code))));
      case STEP_CODE:
	return (make_step_code (compile_lcode(code)));
      case DEFAULT_CODE:
	return NULL;
      }
  } else { /* compressed names for local variables */
    switch SWIT_FIXNUM(c)
      {
      case 0: return make_lvar_code(swit_zero, swit_zero);
      case 1: return make_lvar_code(swit_zero, swit_one);
      case 2: return make_lvar_code(swit_zero, swit_two);
      case 3: return make_lvar_code(swit_zero, swit_three);
      case 4: return make_lvar_code(swit_one, swit_zero);
      case 5: return make_lvar_code(swit_one, swit_one);
      case 6: return make_lvar_code(swit_one, swit_two);
      case 7: return make_lvar_code(swit_two, swit_zero);
      case 8: return make_lvar_code(swit_two, swit_one);
      case 9: return make_lvar_code(swit_two, swit_two);
      }
  }
  return NULL;
}

extern SWiT swit_magic;
extern SWiT_Ccont done_cont;

static SWiT
eval_prim(SWiT exp, SWiT qenv)
{
  if SWIT_VECTORP(exp) {
    SWiT *code = VECTOR_ELS(exp);
    if (*code == swit_magic) {
      return swit_eval(compile_lcode(code[1]), SWIT_ENV(qenv), done_cont);
    } else {
      return exp;
    }
  } else {
    SWiT_Expr args[2];
    SWiT code;
    args[0] = swit_make_quote(exp);
    args[1] = swit_make_quote(qenv);
    if ((code = swit_apply_proc(":compile", 2, args))) {
      return swit_eval(compile_lcode(code), SWIT_ENV(qenv), done_cont);
    } else {
      prim_err_msg = comp_impl_err;
      exn = exn_unknown;
      return NULL;
    }
  }
}

extern SWiT swit_error_object;

SWiT_Expr
read_program (int token)
{
  SWiT datum = rd_datum(token);
  if (datum == NULL) {
    return swit_make_quote(swit_error_object);
  } else {
    SWiT lcode;
    SWiT_Expr args[2];
    args[0] = swit_make_quote(datum);
    args[1] = swit_make_quote(swit_void_env);
    /*   SWiT_Expr e = swit_make_quote(datum);*/
    if ((lcode = swit_apply_proc(":compile", 2, args))) {
      return(compile_lcode(lcode));
    } else {
      fprintf(stderr, comp_impl_err);
      return swit_make_quote(swit_error_object);
    }
  }
}


SWiT_Expr
compile_program (SWiT datum)
{
  if SWIT_VECTORP(datum) {
    /* Already compiled ? (first pass) */
    SWiT *code = VECTOR_ELS(datum);
    if (*code == swit_magic) {
      if SWIT_STRINGP(code[1]) {
	SWiT load_sym = swit_make_symbol("load", 4, swit_unbound);
	SWiT_Expr arg = swit_make_quote(code[1]);
	/* (load "file") */
	return make_call_code(swit_make_quote(SYMBOL_VAL(SWIT_SYM(load_sym))), 1, &arg);
      } else {
	return compile_lcode(code[1]);
      }
    } else {
      return swit_make_quote(datum);
    }
  } else {
    SWiT lcode;
    SWiT_Expr args[2];
    args[0] = swit_make_quote(datum);
    args[1] = swit_make_quote(swit_void_env);

    /*SWiT_Expr e = swit_make_quote(datum);*/
    if ((lcode = swit_apply_proc(":compile", 2, args))) {
      return(compile_lcode(lcode));
    } else {
      fprintf(stderr, comp_impl_err);
      return swit_make_quote(swit_error_object);
    }
  }
}

SWiT_Expr
read_exp (int token, SWiT env)
{
  SWiT code;
  SWiT_Expr args[2];
  SWiT datum = rd_datum(token);
  if (datum == NULL) {
    return swit_make_quote(swit_error_object);
  }
  args[0] = swit_make_quote(datum);
  args[1] = swit_make_quote(env);
  /*  e = swit_make_quote(datum);*/
  if ((code = swit_apply_proc(":compile", 2, args))) {
    return(compile_lcode(code));
  } else {
    fprintf(stderr, comp_impl_err);
    return swit_make_quote(swit_error_object);
  }
}

#define GEN_CODE0(name) \
static SWiT \
mk_ ## name ## _prim ()\
{\
  return (swit_make_pair (inst[name ## _CODE], swit_null_list));\
}

GEN_CODE0(DEFAULT)

#define GEN_CODE1(name) \
static SWiT \
mk_ ## name ## _prim (SWiT expr)\
{\
  return (swit_make_pair (inst[name ## _CODE], expr));\
}

GEN_CODE1(QUOTE)
GEN_CODE1(STEP)
GEN_CODE1(SYMBOL)
GEN_CODE1(AND)
GEN_CODE1(OR)
GEN_CODE1(BEGIN)
GEN_CODE1(ARROW)
GEN_CODE1(DELAY)
GEN_CODE1(UNDEFINE)

#define GEN_CODE2(name) \
static SWiT \
mk_ ## name ## _prim (SWiT arg1, SWiT arg2)\
{\
  return (swit_make_pair (inst[name ## _CODE], \
			  swit_make_pair(arg1, arg2)));\
}


GEN_CODE2(LAMBDA)
GEN_CODE2(DEFMACRO)
GEN_CODE2(DEFINE)
GEN_CODE2(COND)
GEN_CODE2(SET)
GEN_CODE2(LET)
GEN_CODE2(LETSTAR)
GEN_CODE2(LETREC)
GEN_CODE2(PCALL)
GEN_CODE2(ECALL)
GEN_CODE2(SCALL)

static SWiT
mk_LVAR_prim (SWiT arg1, SWiT arg2)\
{
  switch SWIT_FIXNUM(arg1)
    {
    case 0:
      if (SWIT_FIXNUM(arg2) < 4) {
	return arg2;
      } else {
	goto gen;
      }
    case 1:
      if (SWIT_FIXNUM(arg2) < 3) {
	return inst[4+SWIT_FIXNUM(arg2)];
      } else {
	goto gen;
      }
    case 2:
      if (SWIT_FIXNUM(arg2) < 3) {
	return inst[7+SWIT_FIXNUM(arg2)];
      } else {
	goto gen;
      }
    default: goto gen;
    }
gen:
  return (swit_make_pair (inst[LVAR_CODE],
			  swit_make_pair(arg1, arg2)));
}

#define GEN_CODE3(name) \
static SWiT \
mk_ ## name ## _prim (SWiT arg1, SWiT arg2, SWiT arg3)\
{\
  return (swit_make_pair (inst[name ## _CODE], \
			  swit_make_pair(arg1, \
					 swit_make_pair(arg2, arg3))));\
}

GEN_CODE3(IF)
GEN_CODE3(CASE)
GEN_CODE3(TRY)

/* TODO (?)
static SWiT
decode (SWiT_Expr e)
{
  switch (SWIT_INST(e))
    {
    case QUOTE_CODE:
      return (decode_quote(e));
    case PCALL_CODE:
      return (decode_pcall(e));
    case SCALL_CODE:
      return (decode_scall(e));
    case ECALL_CODE:
      return (decode_ecall(e));
    case SYMBOL_CODE:
      return (decode_symbol(e));
    case LVAR_CODE:
      return (decode_lvar(e));
    case LAMBDA_CODE:
      return (decode_lambda(e));
    case IF_CODE:
      return (decode_if(e));
    case TRY_CODE:
      return (decode_try(e));
    case AND_CODE:
      return (decode_and(e));
    case OR_CODE:
      return (decode_or(e));
    case BEGIN_CODE:
      return (decode_begin(e));
    case ARROW_CODE:
      return (decode_arrow(e));
    case COND_CODE:
      return (decode_cond(e));
    case CASE_CODE:
      return (decode_case(e));
    case SET_CODE:
      return (decode_set(e));
    case LET_CODE:
      return (decode_let(e));
    case LETSTAR_CODE:
      return (decode_letstar(e));
    case LETREC_CODE:
      return (decode_letec(e));
    case DELAY_CODE:
      return (decode_delay(e));
    case DEFMACRO_CODE:
      return (decode_defmacro(e));
    case DEFINE_CODE:
      return (decode_define(e));
    case STEP_CODE:
      return (decode_step(e));
    case DEFAULT_CODE:
      return (decode_default(e));
    default:
      swit_signal_error(swit_else, "[DECODE] -- Unknown code");
    }
}

*/


SWiT_Expr swit_nil_fun;

void
swit_init_code (void)
{
  int i;
  SWiT params = swit_make_pair(swit_zero, swit_null_list);
  SWiT_Expr body = swit_make_quote(swit_null_list);
  SWiT code = swit_make_pair
    (swit_make_fixnum(20),
     swit_make_pair
     (params, swit_make_pair(swit_zero, swit_null_list)));
  swit_nil_fun = make_lambda_code(params, body, code);

  for (i = 0; i < 100; i++) inst[i] = swit_make_fixnum(i);
  swit_two = inst[2];
  swit_three = inst[3];

  swit_make_prim0(":code:default", mk_DEFAULT_prim);
  swit_make_prim1(":code:quote", mk_QUOTE_prim,
		  ANY);
  swit_make_prim1(":code:step", mk_STEP_prim,
		  ANY);
  swit_make_prim2(":code:pcall", mk_PCALL_prim,
		  ANY, ANY);
  swit_make_prim2(":code:scall", mk_SCALL_prim,
		  SYMBOL, ANY);
  swit_make_prim2(":code:ecall", mk_ECALL_prim,
		  ANY, ANY);
  swit_make_prim1(":code:symbol", mk_SYMBOL_prim,
		  ANY);
  swit_make_prim1(":code:and", mk_AND_prim,
		  VECT_M);
  swit_make_prim1(":code:or", mk_OR_prim,
		  VECT_M);
  swit_make_prim1(":code:begin", mk_BEGIN_prim,
		  VECT_M);
  swit_make_prim1(":code:arrow", mk_ARROW_prim,
		  ANY);
  swit_make_prim1(":code:delay", mk_DELAY_prim,
		  ANY);
  swit_make_prim2(":code:lvar", mk_LVAR_prim,
		  FIXNUM, FIXNUM);
  swit_make_prim2(":code:let", mk_LET_prim,
		  ANY, ANY);
  swit_make_prim2(":code:let*", mk_LETSTAR_prim,
		  ANY, ANY);
  swit_make_prim2(":code:letrec", mk_LETREC_prim,
		  ANY, ANY);
  swit_make_prim2(":code:set", mk_SET_prim,
		  ANY, ANY);
  swit_make_prim2(":code:cond", mk_COND_prim,
		  VECT_M, VECT_M);
  swit_make_prim2(":code:define", mk_DEFINE_prim,
		  SYMBOL, ANY);
  swit_make_prim3(":code:try", mk_TRY_prim,
		  ANY, ANY, ANY);
  swit_make_prim2(":code:defmacro", mk_DEFMACRO_prim,
		  SYMBOL, ANY);
  swit_make_prim3(":code:if", mk_IF_prim,
		  ANY, ANY, ANY);
  swit_make_prim3(":code:case", mk_CASE_prim,
		  ANY, VECT_M, VECT_M);
  swit_make_prim2(":code:lambda", mk_LAMBDA_prim,
		  ANY, ANY);

  swit_make_prim12 ("eval", eval_prim,
		    ANY, ENV,
		    swit_void_env);
}
