/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public 
License for more details.

You should have received a copy of the GNU Library General Public License 
along with SWiT software; see the file COPYING.  
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/
#include "swit.h"
#include "code.h"

extern SWiT cur_out_port;
extern SWiT swit_print_depth;
extern void print_sym (SWiT_Symbol obj);
extern void prim2buff (SWiT name);
extern void print_pair (SWiT expr, int esc, int tcl, SWiT fv, int depth);
extern void print_pair_std (SWiT obj, int pflag, int esc, int tcl, SWiT fv, int depth);

#define current_op SWIT_PORT(cur_out_port)
#define out_port PORT_DESCR(current_op)
#define print_buffer PORT_BUFFER(current_op)
#define print_cur_idx PORT_INDEX(current_op)
#define print_base_idx PORT_EOB(current_op)
#define write_fun PORT_WRITE_FUN(current_op)

/* exported */
void print_code_to_buffer (SWiT_Expr obj, SWiT fv, int depth);
void print_lambda (SWiT_Lambda lbd, SWiT fv, int depth);
void swit_print_code (SWiT_Expr obj, SWiT fv);


static char *out_err = "output: cannot write"; 
static int print_depth;

/*-------------------------------------------------------------
 * Disassembling functions prototypes (forward declarations) :
 *-------------------------------------------------------------
 */

static void print_body (SWiT_Expr obj, SWiT fv, int depth);
static void print_def (SWiT_Define def_data, SWiT fv, int macro_flag, int depth);
static void print_gset (SWiT_Gset gset, SWiT fv, int depth);
static void print_lset (SWiT_Lset lset, SWiT fv, int depth);
static void print_lvar (int var, SWiT fv);
static void print_call0 (SWiT_Sym_prim0 prim, SWiT fv, int depth);
static void print_call0n (SWiT_Call0n cll, SWiT fv, int depth);
static void print_call1 (SWiT_Call1 cll, SWiT fv, int depth);
static void print_call1n (SWiT_Call1n cll, SWiT fv, int depth);
static void print_call2 (SWiT_Call2 cll, SWiT fv, int depth);
static void print_call3 (SWiT_Call3 cll, SWiT fv, int depth);
static void print_quote (SWiT obj, SWiT fv, int depth);
static void print_delay (SWiT_Expr obj, SWiT fv, int depth);
static void print_if (SWiT_If if_data, SWiT fv, int depth);
static void print_and_or (SWiT_Begin vec, char *init, SWiT fv, int depth);
static void print_case (SWiT_Case data, SWiT fv, int depth);
static void print_cond (SWiT_Cond data, SWiT fv, int depth);
static void print_letstar (SWiT_Let let, SWiT fv, int depth);
static void print_letrec (SWiT_Let letrec, SWiT fv, int depth);
static void print_let (SWiT_Let let, SWiT fv, int depth);
static void print_body (SWiT_Expr obj, SWiT fv, int depth);
static void print_begin (SWiT_Begin beg, SWiT fv, int depth);

/* debug code */
void
debug_code(SWiT_Expr obj)
{
  swit_print_code(obj, swit_null_list);
  fprintf(stderr,"\n");
}

void
swit_print_code (SWiT_Expr obj, SWiT fv)
{
  int cnt, len;
  print_cur_idx = 0;
  print_code_to_buffer(obj, fv, 0);
  len = print_cur_idx;
  if ((cnt = (*write_fun)(out_port, print_buffer, len)) != len) {
    swit_signal_error(swit_else, out_err);
  }
  PORT_EOB(current_op) = 0;
  PORT_INDEX(current_op) = 0;
}

/*---------------------------------
 * Printing code (= disassembling)
 *---------------------------------
 */


void
print_lambda (SWiT_Lambda lbd, SWiT fv , int depth)
{
  SWiT_Expr args[2];
  SWiT datum;
  args[0] = swit_make_quote(LAMBDA_PCODE(lbd));
  args[1] = swit_make_quote(fv);
  if ((datum = swit_apply_proc(":pcode->datum", 2, args))) {
    print_pair_std(datum, 1, 1, 0, fv, depth);
  } else {
    fprintf(stderr, "error in :pcode->datum");
  }
  /*  print_pair_std(LAMBDA_PCODE(lbd), 1, 1, 0, fv, depth);*/
}

static void
print_begin (SWiT_Begin beg, SWiT fv, int depth)
{
  int i,size = BEG_NFORMS(beg);
  SWiT_Expr *forms = BEG_FORMS(beg);

  depth++;
  char2buff('(');
  str2buff("begin");
  for (i = 0; i < size; i++) {
    space2buff();
    print_code_to_buffer(forms[i], fv, depth);
  }
  char2buff(')');
}

static void
print_car_names (SWiT name_list)
{
  if (SWIT_PAIRP(name_list) && SWIT_PAIRP(SWIT_CAR(name_list))
      && SWIT_SYMBOLP(SWIT_CAR(SWIT_CAR(name_list)))) {
    print_sym(SWIT_SYM(SWIT_CAR(SWIT_CAR(name_list))));
  } else {
    str2buff("???");
  }
}

static void
print_body (SWiT_Expr obj, SWiT fv, int depth)
{
  space2buff();
  if (SWIT_INST(obj) == BEGIN_CODE) {
    print_begin(SWIT_BEGIN(obj), fv, depth);
  } else {
    print_code_to_buffer(obj, fv, depth);
  }
}

static void
print_let (SWiT_Let let, SWiT fv, int depth)
{
  SWiT_Lambda lbd = LET_LBD(let);
  SWiT_Expr *args = LET_ARGS(let);
  SWiT names = LAMBDA_NAMES(lbd), saved;
  int arity = LAMBDA_NPARAMS(lbd),i;

  depth++;
  saved = names;
  char2buff('(');
  str2buff("let ");
  char2buff('(');
  for (i = 0; i < arity; ++i) {
    char2buff('(');
    print_car_names(names);
    space2buff();
    print_code_to_buffer(args[i], fv, depth);
    names = SWIT_CDR(names);
    char2buff(')');
    if (i != arity - 1) {
      space2buff();
    }
  }
  char2buff(')');
  depth++;
  print_body(LAMBDA_BODY(lbd), swit_make_pair(saved, fv), depth);
  char2buff(')');
}

static void
print_letrec (SWiT_Let letrec, SWiT fv, int depth)
{
  SWiT_Lambda lbd = LET_LBD(letrec);
  SWiT_Expr *args = LET_ARGS(letrec);
  SWiT names = LAMBDA_NAMES(lbd);
  int arity = LAMBDA_NPARAMS(lbd),i;

  depth++;
  fv = swit_make_pair(names, fv);
  char2buff('(');
  str2buff("letrec ");
  char2buff('(');
  for (i = 0; i < arity; ++i) {
    char2buff('(');
    print_car_names(names);
    space2buff();
    print_code_to_buffer(args[i], fv, depth);
    names = SWIT_CDR(names);
    char2buff(')');
    if (i != arity-1) {
      space2buff();
    }
  }
  char2buff(')');
  depth++;
  print_body(LAMBDA_BODY(lbd), fv, depth);
  char2buff(')');
}

static void
print_letstar (SWiT_Let let, SWiT fv, int depth)
{
  SWiT_Lambda lbd = LET_LBD(let);
  SWiT_Expr *args = LET_ARGS(let);
  SWiT names = LAMBDA_NAMES(lbd);
  int arity = LAMBDA_NPARAMS(lbd),i;

  depth++;
  char2buff('(');
  str2buff("let* ");
  char2buff('(');
  for (i = 0; i < arity; ++i) {
    SWiT name = swit_make_pair(SWIT_CAR(names), swit_null_list);
    char2buff('(');
    print_car_names(name);
    space2buff();
    print_code_to_buffer(args[i], fv, depth);
    names = SWIT_CDR(names);
    fv = swit_make_pair(name, fv);
    char2buff(')');
    if (i != arity-1) {
      space2buff();
    }
  }
  char2buff(')');
  depth++;
  print_body(LAMBDA_BODY(lbd), fv, depth);
  char2buff(')');
}


static void
print_and_or (SWiT_Begin vec, char *init, SWiT fv, int depth)
{
  SWiT_Expr *els = BEG_FORMS(vec);
  int size = BEG_NFORMS(vec);

  depth++;
  str2buff(init);
  if (size == 0) {
    char2buff(')');
  } else {
    int i;
    print_code_to_buffer(els[0], fv, depth);
    for (i = 1; i < size; ++i) {
      space2buff();
      print_code_to_buffer(els[i], fv, depth);
    }
    char2buff(')');
  }
}

static void
print_arrow (SWiT_Expr expr, SWiT fv, int depth)
{
  str2buff("=> ");
  print_code_to_buffer(expr, fv, depth);
}


static void
print_cond (SWiT_Cond data, SWiT fv, int depth)
{
  SWiT_Expr *tests = COND_TESTS(data);
  SWiT_Expr *seqs = COND_SEQS(data);
  int i, nclauses = COND_NCLAUSES (data);

  depth++;
  char2buff('(');
  str2buff("cond ");
  depth++;
  for (i = 0; i < nclauses-1; i++) {
    char2buff('(');
    print_code_to_buffer(tests[i], fv, depth);
    if (seqs[i]) {
      print_body(seqs[i], fv, depth);
    }
    char2buff(')');
    space2buff();
  }
  char2buff('(');
  if (SWIT_QUOTE(tests[i]) == swit_true) {
    str2buff("else");
  } else {
    print_code_to_buffer(tests[i], fv, depth);
  }
  if (seqs[i]) {
    print_body(seqs[i], fv, depth);
  }
  char2buff(')');
  char2buff(')');
}

static void
print_case (SWiT_Case data, SWiT fv, int depth)
{
  SWiT_Expr key = CASE_KEY(data);
  SWiT *tests = CASE_TESTS(data);
  SWiT_Expr *seqs = CASE_SEQS(data);
  int i, nclauses = CASE_NCLAUSES (data);

  depth++;
  char2buff('(');
  str2buff("case ");
  print_code_to_buffer(key, fv, depth);
  space2buff();
  depth++;
  for (i = 0; i < nclauses-1; i++) {
    char2buff('(');
    print_pair(tests[i], 1, 0, fv, depth);
    print_body(seqs[i], fv, depth);
    char2buff(')');
    space2buff();
  }
  char2buff('(');
  if (tests[i] == swit_else) {
    str2buff("else");
  } else {
    print_pair(tests[i], 1, 0, fv, depth);
  }
  print_body(seqs[i], fv, depth);
  char2buff(')');
  char2buff(')');
}

static void
print_if (SWiT_If if_data, SWiT fv, int depth)
{
  SWiT el = SWIT_QUOTE(IF_ELSE(if_data));
  depth++;
  char2buff('(');
  if (el == swit_false) {
    str2buff("and ");
  } else {  
    str2buff("if ");
  }
  print_code_to_buffer(IF_TEST(if_data), fv, depth);
  space2buff();
  print_code_to_buffer(IF_THEN(if_data), fv, depth);
  if (el != swit_false) {
    space2buff();
    print_code_to_buffer(IF_ELSE(if_data), fv, depth);
  }
  char2buff(')');
}

#define SWIT_NLITERALP(obj) (SWIT_SYMBOLP(obj) || SWIT_PAIRP(obj))

static void
print_quote (SWiT obj, SWiT fv, int depth)
{
  depth++;
  if (SWIT_NLITERALP(obj)) {
    char2buff('\'');
  }
  print_to_buffer(obj, 1, 0, fv, depth);
}

static void
print_delay (SWiT_Expr obj, SWiT fv, int depth)
{
  depth++;
  char2buff('(');
  str2buff("delay ");
  print_code_to_buffer(obj, fv, depth);
  char2buff(')');
}

/*
static void
print_splice (SWiT_Expr obj, SWiT fv, int depth)
{
  depth++;
  str2buff(",@");
  print_code_to_buffer(obj, fv, depth);
}

static void
print_unquote (SWiT_Expr obj, SWiT fv, int depth)
{
  depth++;
  char2buff(',');
  print_code_to_buffer(obj, fv, depth);
}
*/

static void
print_call (SWiT_Call cll, SWiT fv, int depth)
{
  SWiT_Expr *args = CALL_ARGS(cll);
  int nargs = CALL_NARGS(cll),i;

  depth++;
  char2buff('(');
  print_code_to_buffer(CALL_PROC(cll), fv, depth);
  for (i = 0; i < nargs; ++i) {
    space2buff();
    print_code_to_buffer(args[i], fv, depth);
  }
  char2buff(')');
}

static void
print_call0 (SWiT_Sym_prim0 prim, SWiT fv, int depth)
{
  depth++;
  char2buff('(');
  str2buff("#'");
  prim2buff(PRIM_NAME(prim));
  char2buff(')');
}

static void
print_call0n (SWiT_Call0n cll, SWiT fv, int depth)
{
  int nargs,i;
  SWiT_Expr *args;

  depth++;
  char2buff('(');
  str2buff("#'");
  prim2buff(CALL_NAME(cll));
  nargs = CALL_NARGS(cll);
  args = CALL_ARGS(cll);
  for (i = 0; i < nargs; ++i) {
    space2buff();
    print_code_to_buffer(args[i], fv, depth);
  }
  char2buff(')');
}

#define CALL_VAL(obj) PRIM_VAL(CALL_PRIM(obj))

static void
print_call1 (SWiT_Call1 cll, SWiT fv, int depth)
{
  SWiT prim_name = CALL_NAME(cll);
  SWiT_Expr arg;

  depth++;
  char2buff('(');
  str2buff("#'");
  prim2buff(prim_name);
  arg = CALL_ARG(cll);
  if ((SWIT_INST(arg) == QUOTE_CODE) && (SWIT_QUOTE(arg) == CALL_VAL(cll))) {
    char2buff(')');
  } else {
    space2buff();
    print_code_to_buffer(arg, fv, depth);
  }
}

static void
print_call1n (SWiT_Call1n cll, SWiT fv, int depth)
{
  int nargs,i;
  SWiT_Expr *args;

  depth++;
  char2buff('(');
  str2buff("#'");
  prim2buff(CALL_NAME(cll));
  space2buff();
  print_code_to_buffer(CALL_ARG(cll), fv, depth);
  nargs = CALL_NARGS(cll);
  args = CALL_ARGS(cll);
  for (i = 0; i < nargs; ++i) {
    space2buff();
    print_code_to_buffer(args[i], fv, depth);
  }
  char2buff(')');
}

static void
print_call2 (SWiT_Call2 cll, SWiT fv, int depth)
{
  SWiT prim_name = CALL_NAME(cll);
  SWiT_Expr arg2;

  depth++;
  char2buff('(');
  str2buff("#'");
  prim2buff(prim_name);
  space2buff();
  print_code_to_buffer(CALL_ARG1(cll), fv, depth);
  arg2 = CALL_ARG2(cll);
  if ((SWIT_INST(arg2) == QUOTE_CODE) && (SWIT_QUOTE(arg2) == CALL_VAL(cll))) {
    char2buff(')');
  } else {
    space2buff();
    print_code_to_buffer(arg2, fv, depth);
  }
}

static void
print_call2n (SWiT_Call2n cll, SWiT fv, int depth)
{
  int nargs,i;
  SWiT_Expr *args;

  depth++;
  char2buff('(');
  str2buff("#'");
  prim2buff(CALL_NAME(cll));
  space2buff();
  print_code_to_buffer(CALL_ARG1(cll), fv, depth);
  space2buff();
  print_code_to_buffer(CALL_ARG2(cll), fv, depth);
  nargs = CALL_NARGS(cll);
  args = CALL_ARGS(cll);
  for (i = 0; i < nargs; ++i) {
    space2buff();
    print_code_to_buffer(args[i], fv, depth);
  }
  char2buff(')');
}

static void
print_call3 (SWiT_Call3 cll, SWiT fv, int depth)
{
  depth++;
  char2buff('(');
  str2buff("#'");
  prim2buff(CALL_NAME(cll));
  space2buff();
  print_code_to_buffer(CALL_ARG1(cll), fv, depth);
  space2buff();
  print_code_to_buffer(CALL_ARG2(cll), fv, depth);
  space2buff();
  print_code_to_buffer(CALL_ARG3(cll), fv, depth);
  char2buff(')');
}

static void
print_lvar (int var, SWiT fv)
{
  char buf[7];
  int i;
  int db = VAR_DB(var);
  int idx = VAR_IDX(var);
  for (i = 0; i < db; i++) {
    if (SWIT_PAIRP(fv)) {
      fv = SWIT_CDR(fv);
    } else {
      sprintf(buf,"%d+%d", db, idx);
      str2buff(buf);
      return;
    }
  }
  if (SWIT_PAIRP(fv)) {
    fv = SWIT_CAR(fv);
  } else { /* for debugging only */
    sprintf(buf,"%d+%d", db, idx);
    str2buff(buf);
    return;
  }
  for (i = 0; i < idx; i++) {
    if (SWIT_PAIRP(fv)) {
      fv = SWIT_CDR(fv);
    } else {  /* for debugging only */
      sprintf(buf,"%d+%d", db, idx);
      str2buff(buf);
      return;
    }
  }
  print_car_names(fv);
}

static void
print_lset (SWiT_Lset lset, SWiT fv, int depth)
{
  int var = LSET_VAR(lset);

  depth++;
  char2buff('(');
  str2buff("set! ");
  print_lvar(var, fv);
  space2buff();
  print_code_to_buffer(LSET_EXPR(lset), fv, depth);
  char2buff(')');
}

static void
print_gset (SWiT_Gset gset, SWiT fv, int depth)
{
  depth++;
  char2buff('(');
  str2buff("set! ");
  print_sym(GSET_SYM(gset));
  space2buff();
  print_code_to_buffer(GSET_EXPR(gset), fv, depth);
  char2buff(')');
}

static void
print_def (SWiT_Define def_data, SWiT fv, int macro_flag, int depth)
{
  depth++;
  char2buff('(');
  if (macro_flag)
    str2buff("defmacro ");
  else
    str2buff("define ");
  print_to_buffer(DEFINE_SYM(def_data), 1, 0, fv, depth);
  space2buff();
  print_code_to_buffer(DEFINE_EXPR(def_data), fv, depth);
  char2buff(')');
}

static void
print_try (SWiT_Try try, SWiT fv, int depth)
{
  SWiT exn_lst = TRY_EXN(try);
  SWiT_Expr *handlers = TRY_HDL(try);
  str2buff("(try ");
  print_code_to_buffer(TRY_EXPR(try), fv, depth);
  while SWIT_PAIRP(exn_lst) {
    str2buff(" (");
    print_to_buffer(SWIT_CAR(exn_lst), 1, 0, fv, depth);
    space2buff();
    print_code_to_buffer(*handlers, fv, depth);
    char2buff(')');
    exn_lst = SWIT_CDR(exn_lst);
    handlers++;
  }
  char2buff(')');
}

void
print_code_to_buffer (SWiT_Expr obj, SWiT fv, int depth)
{
  print_depth = SWIT_FIXNUM(SYMBOL_VAL(SWIT_SYM(swit_print_depth)));
  if (depth > 10*print_depth) {
    str2buff("..");
  } else {
    switch (SWIT_INST(obj))
      {
      case LAMBDA_CODE: 
	{
	  SWiT_Expr args[2];
	  SWiT datum;
	  args[0] = swit_make_quote(LAMBDA_PCODE(SWIT_LAMBDA(obj)));
	  args[1] = swit_make_quote(fv);
	  if ((datum = swit_apply_proc(":pcode->datum", 2, args))) {
	    print_pair_std(datum, 1, 1, 0, fv, depth);
	  } else {
	    fprintf(stderr, "error in :pcode->datum");
	  }
	  /*print_pair_std(LAMBDA_PCODE(SWIT_LAMBDA(obj)), 1, 1, 0, fv, depth);*/
	  break;
	}
      case LVAR_CODE:
	print_lvar(SWIT_VAR(obj), fv); break;
      case SYMBOL_CODE:
	print_sym(SWIT_SYMBOL(obj)); break;
      case IF_CODE:
	print_if(SWIT_IF(obj), fv, depth); break;
      case AND_CODE:
	print_and_or(SWIT_BEGIN(obj),"(and ", fv, depth); break;
      case OR_CODE:
	print_and_or(SWIT_BEGIN(obj),"(or ", fv, depth); break;
      case COND_CODE:
	print_cond(SWIT_COND(obj), fv, depth); break;
      case CASE_CODE:
	print_case(SWIT_CASE(obj), fv, depth); break;
      case LET_CODE:
	print_let(SWIT_LET(obj), fv, depth); break;
      case LETREC_CODE:
	print_letrec(SWIT_LET(obj), fv, depth); break;
      case LETSTAR_CODE:
	print_letstar(SWIT_LET(obj), fv, depth); break;
      case GSET_CODE:
	print_gset(SWIT_GSET(obj), fv, depth); break;
      case BEGIN_CODE:
	print_begin(SWIT_BEGIN(obj), fv, depth); break;
      case LSET_CODE:
	print_lset(SWIT_LSET(obj), fv, depth); break;
      case QUOTE_CODE:
      case PROTECT_CODE:
	print_quote((SWiT)SWIT_CODE(obj), fv, depth); break;
      case DELAY_CODE:
	print_delay((SWiT_Expr)SWIT_CODE(obj), fv, depth); break;
      case ELSE_CODE:
	str2buff("else "); break;
      case TRY_CODE:
	print_try((SWiT_Try)SWIT_CODE(obj), fv, depth); break;
      case ARROW_CODE:
	print_arrow((SWiT_Expr)SWIT_CODE(obj), fv, depth); break;
      case STEP_CODE:
	print_code_to_buffer((SWiT_Expr)SWIT_CODE(obj), fv, depth); break;
      case CALL_CODE:
	print_call(SWIT_CALL(obj), fv, depth); break;
      case CALL0_CODE:
	print_call0(SWIT_CALL0(obj), fv, depth); break;
      case THUNK_CODE:
	print_call(SWIT_CALL(obj), fv, depth); break;
      case CALL0N_CODE:
	print_call0n(SWIT_CALL0n(obj), fv, depth); break;
      case CALL1N_CODE:
	print_call1n(SWIT_CALL1n(obj), fv, depth); break;
      case CALL2N_CODE:
	print_call2n(SWIT_CALL2n(obj), fv, depth); break;
      case CALL1_CODE:
	print_call1(SWIT_CALL1(obj), fv, depth); break;
      case CALL2_CODE:
	print_call2(SWIT_CALL2(obj), fv, depth); break;
      case CALL3_CODE:
	print_call3(SWIT_CALL3(obj), fv, depth); break;
      case DEFINE_CODE:
	print_def(SWIT_DEFINE(obj), fv, 0, depth); break;
      case DEFMACRO_CODE:
	print_def(SWIT_DEFINE(obj), fv, 1, depth); break;
      default: print_to_buffer((SWiT)obj, 1, 0, fv, depth);
      }
  }
}

/*==================================================================
 *                       PRIMITIVE(S)
 *==================================================================
 */

extern SWiT make_fv(SWiT_Env e);
static SWiT
disassemble_prim (SWiT arg)
{
  if SWIT_CLOSUREP(arg) {
    SWiT_Closure clos = SWIT_CLOSURE(arg);
    SWiT_Lambda lbd = CLOSURE_LBD(clos);
    SWiT_Env env = CLOSURE_ENV(clos);
    SWiT_Expr args[2];
    SWiT res;
    args[0] = swit_make_quote(LAMBDA_PCODE(lbd));
    /* env not disassembled for the moment */
    args[1] = swit_make_quote(make_fv(env));
    if ((res = swit_apply_proc(":pcode->datum", 2, args))) {
      return res;
    } else {
      exn = exn_unknown;
      prim_err_msg = dis_impl_err;
      return NULL;
    }
    /*return LAMBDA_PCODE(lbd);*/
  } else {
    return arg;
  }
}

/*==================================================================
 *                       INITIALISATION
 *==================================================================
 */

void
swit_init_disassemble (void)
{
  swit_make_prim1("disassemble", disassemble_prim,
		  ANY);
}
