/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "swit.h"
#include "code.h"

#define APPLY0(op) (PRIM_FUN(op)())
#define APPLY1(op,x) (PRIM_FUN(op)(x))
#define APPLY2(op,x,y) (PRIM_FUN(op)(x,y))
#define APPLY3(op,x,y,z) (PRIM_FUN(op)(x,y,z))
#define APPLY4(op,x,y,z,t) (PRIM_FUN(op)(x,y,z,t))

/* imported */
extern int swit_eqv (SWiT obj1, SWiT obj2);
extern SWiT swit_list_to_vector (SWiT arg);
extern SWiT swit_append2 (SWiT lst1, SWiT lst2);
extern void swit_print_code (SWiT_Expr obj, SWiT fv);
extern SWiT make_fv(SWiT_Env e);
extern int search (SWiT sym, SWiT names);
extern SWiT usr2_hook; /* defined in switMisc.c */
extern int usr2;  /* defined in switTop.c */
extern SWiT swit_make_env (SWiT_Env env);
extern int issocket(int des);
extern SWiT make_fv(SWiT_Env e);
extern void swit_printl (SWiT obj, SWiT fv);
extern int read_paren;
extern int prompt;
extern char *filename;
extern SWiT swit_prompt;
extern SWiT swit_break_prompt;
extern SWiT cur_in_port;
extern int step_start;
extern int transcript_descr;
extern int tk;

/* exported everywhere */
char *prim_err_msg = "";

/* exported to switMisc.c and switControl.c */
SWiT_Env
swit_extend_env (SWiT names, SWiT *vals, SWiT_Env env)
{
  SWiT_Env frame = SWIT_ALLOC(Env);
  ENV_FRAME(frame) = vals;
  ENV_NAMES(frame) = names;
  ENV_NEXT(frame) = env;
  return (frame);
}

static int
env_depth (SWiT_Env env)
{
  int cnt = 0;
  while (env) {
    cnt++;
    env = ENV_NEXT(env);
  }
  return cnt;
}
static SWiT_Ccont
extend_cont (SWiT_Env env, SWiT_Expr unev, void *vals, SWiT_Ccont cont)
{
  SWiT_Ccont frame = SWIT_ALLOC(Ccont);
  CCONT_ENV(frame) = env;
  CCONT_UNEV(frame) = unev;
  CCONT_VALS(frame) = vals;
  CCONT_NEXT(frame) = cont;
  return (frame);
}

static SWiT
make_cont (SWiT_Ccont cont)
{
  SWiT obj = swit_alloc_object();
  SWIT_TYPE(obj) = CONT;
  SWIT_CONT(obj) = cont;
  return obj;
}

/* exported to switInit.c */
SWiT_Ccont
init_cont (int inst)
{
  SWiT_Expr expr = swit_alloc_expr();
  SWIT_INST(expr) = inst;
  return extend_cont(NULL, expr, NULL, NULL);
}


SWiT_Exn exn_stack;

static void
push_exn (SWiT_Try try, SWiT_Env env, SWiT_Ccont cont)
{
  SWiT_Exn cell = (struct SWiT_Exn *)swit_malloc(sizeof(struct SWiT_Exn));
  cell->try = try;
  cell->env = env;
  cell->cont = cont;
  cell->next = exn_stack;
  exn_stack = cell;
}

/* for cerror */
static SWiT
mk_sym(SWiT_Symbol sym)
{
  SWiT val = swit_alloc_object();
  SWIT_TYPE(val) = SYMBOL;
  SWIT_SYM(val) = sym;
  return val;
}

/* reinitialized in switTop.c */
int step_start;

extern SWiT_Ccont done_cont;
extern SWiT_Ccont break_cont;


SWiT
swit_eval (SWiT_Expr obj, SWiT_Env env, SWiT_Ccont ccont)
{
  SWiT val;
  SWiT *args;
  SWiT fun;
  int nargs = 0; /* pour viter le warning */

  /* On calcule la valeur de "obj" dans l'environnement "env", on
     met le rsultat dans "val", puis on va  "next".
     Si le calcul de "obj" ncessite le calcul de sous-expressions,
     on raffecte "obj" pour calculer la premire sous-expression,
     et on augmente la continuation courante de ce qu'il y reste  faire
     quand la valeur de cette premire ss-expr aura t obtenue.
  */
 eval:
  switch SWIT_INST(obj)
    {
    case STEP_CODE:
      {
	SWiT_Expr expr = (SWiT_Expr)SWIT_CODE(obj);
	static SWiT history;
	static int skip;
	if (step_start) {
	  skip = 0;
	  step_start = 0;
	  history = swit_null_list;
	}
	if (skip) {
	  obj = expr;
	  goto eval;
	} else {
	  SWiT_Expr r_arrow;
	  SWiT e = make_fv(env);
	  SWiT_Expr un = swit_make_quote(swit_make_fixnum(env_depth(env)));
	  r_arrow = swit_make_quote(swit_apply_proc(":right-arrow", 1, &un));
	  swit_apply_proc("display", 1, &r_arrow);
	  swit_print_code(expr, e);
	  history = swit_make_pair(swit_make_pair((SWiT)expr, e), history);
	  switch (getchar()) {
	  case 'b': /* show history */
	    {
	      SWiT h = history;
	      while (h != swit_null_list) {
		swit_print_code((SWiT_Expr)SWIT_CAR(SWIT_CAR(h)), SWIT_CDR(SWIT_CAR(h)));
		fprintf(stderr,"\n");
		h = SWIT_CDR(h);
	      }
	      break;
	    }
	  case 'q':case 'c':
	    skip = 1;break;
	  }
	  ccont = extend_cont(NULL, obj, NULL, ccont);
	  obj = expr;
	  goto eval;
	}
      }
    case QUOTE_CODE:
      {
	val = SWIT_QUOTE(obj);
	goto next;
      }
    case IF_CODE:
      {
	ccont = extend_cont(env, obj, NULL, ccont);
	obj = IF_TEST(SWIT_IF(obj));
	goto eval;
      }
    case BEGIN_CODE:
    case OR_CODE:
    case AND_CODE:
      {
	SWiT_Begin beg = SWIT_BEGIN(obj);
       	ccont = extend_cont(env, obj, (void *)BEG_NFORMS(beg), ccont);
	obj = BEG_FORMS(beg)[0];
	goto eval;
      }
    case DEFMACRO_CODE:
    case DEFINE_CODE:
      {
	SWiT_Define def = SWIT_DEFINE(obj);
	ccont = extend_cont(NULL, obj, SWIT_SYM(DEFINE_SYM(def)), ccont);
	obj = DEFINE_EXPR(def);
	goto eval;
      }
    case GSET_CODE:
      {
	SWiT_Gset gset = SWIT_GSET(obj);
	SWiT_Symbol sym = GSET_SYM(gset);

	if (SYMBOL_VAL(sym) == swit_unbound) {
	  prim_err_msg = isunbound;
	  exn = exn_unbound;
	  exn_val = mk_sym(sym);
	  goto cerror;
	} else {
	  ccont = extend_cont(NULL, obj, sym, ccont);
	  obj = GSET_EXPR(gset);
	  goto eval;
	}
      }
    case SYMBOL_CODE:
      {
	/* env is ignored */
	SWiT_Symbol sym = SWIT_SYMBOL(obj);
	val = SYMBOL_VAL(sym);
	if (val == swit_unbound) {
	  prim_err_msg = isunbound;
	  exn = exn_unbound;
	  exn_val = mk_sym(sym);
	  goto cerror;
	}
	goto next;
      }
    case LAMBDA_CODE:
      {
	SWiT_Lambda lbd = SWIT_LAMBDA(obj);
	SWiT_Closure cl = SWIT_ALLOC(Closure);
	CLOSURE_LBD(cl) = lbd;
	CLOSURE_ENV(cl) = env;
	val = swit_alloc_object();
	SWIT_TYPE(val) = CLOSURE;
	SWIT_CLOSURE(val) = cl;
	goto next;
      }
    case LVAR_CODE:
      {
	unsigned int var = SWIT_VAR(obj);
	register int i;
	int db = VAR_DB(var);
	SWiT_Env e = env;
	for (i = 0; i < db; i++) {
	  e = ENV_NEXT(e);
	}
	val = ENV_FRAME(e)[VAR_IDX(var)];
	goto next;
      }
    case LSET_CODE:
      {
	SWiT_Lset lset = SWIT_LSET(obj);
	unsigned int var = LSET_VAR(lset);
	register int i;
	int db = VAR_DB(var);
	SWiT_Env e = env;
	for (i = 0; i < db; i++) {
	  e = ENV_NEXT(e);
	}
	ccont = extend_cont(NULL, obj, ENV_FRAME(e)+VAR_IDX(var), ccont);
	obj = LSET_EXPR(lset);
	goto eval;
      }
    case DELAY_CODE:
      {
	SWiT_Promise prom = SWIT_ALLOC(Promise);
	PROMISE_VAL(prom) = NULL;
	PROMISE_EXPR(prom) = (SWiT_Expr)SWIT_CODE(obj) ;
	PROMISE_ENV(prom) = env;
	val = swit_alloc_object();
	SWIT_TYPE(val) = PROMISE;
	SWIT_PROMISE(val) = prom;
	goto next;
      }
    case THUNK_CODE:
      {
	SWiT_Call cll = SWIT_CALL(obj);
	ccont =  extend_cont(NULL, obj, NULL, ccont);
	obj = CALL_PROC(cll);
	goto eval;
      }
    case CALL0_CODE: /* call prim0 */
      {
	val = APPLY0(SWIT_CALL0(obj));
	goto next;
      }
    case CALL1_CODE: /* call prim1 */
      {
	ccont =  extend_cont(NULL, obj, NULL, ccont);
	obj = CALL_ARG(SWIT_CALL1(obj));
	goto eval;
      }
    case CALL_CODE:
      {
	SWiT_Call cll = SWIT_CALL(obj);
	int n = CALL_NARGS(cll);
	SWiT *vals = (SWiT *)swit_malloc(sizeof(int)*(n+3));
	vals[0] = (SWiT)0;
	vals[1] = (SWiT)n;
	ccont =  extend_cont(env, obj, vals, ccont);
	obj = CALL_PROC(cll);
	goto eval;
      }
    case TRY_CODE:
      {
	SWiT_Try try = SWIT_TRY(obj);
	push_exn(try, env, ccont); /* push new handlers */
	ccont = extend_cont(NULL, obj, NULL, ccont);
	obj = TRY_EXPR(try);
	goto eval;
      }

    default:
      val = (SWiT)obj;
      goto next;
    }

  /* Une valeur a t calcule : on dcide ce que l'on doit faire avec.
     Plusieurs cas :
      1) tous les lments de l'expression "unev" prsente dans
         la continuation courante sont calculs : on termine le calcul
	 (soit set!, soit apply) et on reboucle ici avec ccont_next.
      2) "unev" n'est pas encore compltement calcul, on stoke "val"
         dans l'emplacement prvu de "ccont" et on calcule l'expression
	 suivante (extraite de "unev") dans l'environnement (lui aussi
	 sauvegard dans "ccont") par un goto eval.
      3) "val" est la "dernire" valeur (et la continuation est
         en fait le top-level : A FAIRE ici, ce serait mieux...
	 en effet, le top-level est une continuation particulire)
	 En fait, ce cas ne devrait jamais se produire: il y a tjs
	 une continuation  cause de la boucle sans fin du top-level.

   */
 next:
  {
    SWiT_Expr unev = CCONT_UNEV(ccont);
    switch SWIT_INST(unev)
      {
      case STEP_CODE: {
	SWiT_Expr un = swit_make_quote(swit_make_fixnum(env_depth(env)));
	SWiT_Expr l_arrow = swit_make_quote(swit_apply_proc(":left-arrow", 1, &un));
	swit_apply_proc("display", 1, &l_arrow);
	debug(val);
	ccont = CCONT_NEXT(ccont);
	goto next;
      }
      case DONE:
	{
	  /* plus rien  faire: on retourne "val" */
	  return val;
	}
      case BREAK:
	{
	  swit_printl(val, make_fv(env));
	  printf ("\n");
	  goto break_loop;
	}
      case IF_CODE:
	{
	  if (val == swit_false) {
	    obj = IF_ELSE(SWIT_IF(unev));
	  } else {
	    obj = IF_THEN(SWIT_IF(unev));
	  }
	  env = CCONT_ENV(ccont);
	  ccont = CCONT_NEXT(ccont);
	  goto eval;
	}
      case AND_CODE:
	{
	  if (val == swit_false) {
	    ccont = CCONT_NEXT(ccont);
	    goto next;
	  } else {
	    int n = (int)CCONT_VALS(ccont);
	    obj = BEG_FORMS(SWIT_BEGIN(unev))[n];
	    env = CCONT_ENV(ccont);
	    if (n == 1) {
	      ccont = CCONT_NEXT(ccont);
	    } else {
	      ((int)CCONT_VALS(ccont))--;
	    }
	    goto eval;
	  }
	}
      case OR_CODE:
	{
	  if (val != swit_false) {
	    ccont = CCONT_NEXT(ccont);
	    goto next;
	  } else {
	    int n = (int)CCONT_VALS(ccont);
	    obj = BEG_FORMS(SWIT_BEGIN(unev))[n];
	    env = CCONT_ENV(ccont);
	    if (n == 1) {
	      ccont = CCONT_NEXT(ccont);
	    } else {
	      ((int)CCONT_VALS(ccont))--;
	    }
	    goto eval;
	  }
	}
      case BEGIN_CODE:
	{
	  int n = (int)CCONT_VALS(ccont);
	  obj = BEG_FORMS(SWIT_BEGIN(unev))[n];
	  env = CCONT_ENV(ccont);
	  if (n == 1) {
	    ccont = CCONT_NEXT(ccont);
	  } else {
	    ((int)CCONT_VALS(ccont))--;
	  }
	  goto eval;
	}
      case DEFINE_CODE:
      case DEFMACRO_CODE:
	{
	  SYMBOL_VAL((struct SWiT_Symbol *)CCONT_VALS(ccont)) = val;
	  ccont = CCONT_NEXT(ccont);
	  val = DEFINE_SYM(SWIT_DEFINE(unev));
	  goto next;
	}
      case GSET_CODE:
	{
	  SYMBOL_VAL((struct SWiT_Symbol *)CCONT_VALS(ccont)) = val;
	  ccont = CCONT_NEXT(ccont);
	  goto next;
	}
      case LSET_CODE:
	{
	  *(SWiT *)CCONT_VALS(ccont) = val;
	  ccont = CCONT_NEXT(ccont);
	  goto next;
	}
      case THUNK_CODE:
	{
	  fun = val;
	  nargs = 0;
	  switch SWIT_TYPE(fun) {
	  case PRIM0: goto thunk_prim0;
	  case PRIM01: goto thunk_prim01;
	  case PRIM0N: goto thunk_prim0n;
	  case CLOSURE: goto thunk_closure;
	  case CURRENT_ENV_PRIM: {
	    val = env ? swit_make_env(env) : swit_void_env;
	    ccont = CCONT_NEXT(ccont);
	    goto next;
	  }
	  default:
	    prim_err_msg = not_a_thunk;
	    exn = exn_type;
	    exn_val = val;
	    goto cerror;
	  }
	}
      case CALL1_CODE: {
	SWiT_Sym_prim1 prim = CALL_PRIM(SWIT_CALL1(unev));
	/* val est du bon type (type checked.. ou ANY) */
	val = APPLY1(prim, val); /* pas d'erreur garantie */
	ccont = CCONT_NEXT(ccont);
	goto next;
      }
      case CALL_CODE:
	{
	  SWiT *vals = (SWiT *)CCONT_VALS(ccont);
	  int n = (int)(vals[0]);
	  nargs = (int)(vals[1]);
	  vals[n+2] = val;
	  if (n == nargs) {
	    fun = vals[2];
	    args = vals+3;
	    goto apply;
	  } else {
	    vals[0] = (SWiT)(n+1);
	    obj = CALL_ARGS(SWIT_CALL(unev))[n];
	    env = CCONT_ENV(ccont);
	    goto eval;
	  }
	}
      case PROTECT_CODE:
	/* terminaison  la suite d'une exception */
	/* y a-t-il un "protect" ? */
	/* A FAIRE */
	if SWIT_PROCP(val) {
	  fun = val;
	  args = (SWiT *)CCONT_VALS(ccont);
	  nargs = 1;
	  goto apply;
	} else {
	  ccont = CCONT_NEXT(ccont);
	  goto next;
	}
      case TRY_CODE:
	/* terminaison normale du try */
	/* y a-t-il un "protect" ? */
	/* A FAIRE */
	exn_stack = exn_stack->next; /* pop_exn */
	ccont = CCONT_NEXT(ccont);
	goto next;
      default: goto erreur;
      }
  }

    /* Pourrait tre mis dans "next" dont on termine ici le
       calcul dans la cas particulier du call.
       Trois cas suivant la nature de "fun" :
       1) primitive : on calcul prim(val,...) et on retourne  "next"
       on doit traiter ici des cas particuliers : call/cc
       et apply (ainsi que les fonctionnelles map, for-each etc)
       Pour call/cc, on doit passer la continuation courante
        la fonction argument.
       Le cas de call/cc se subdivise en 3 sous-cas suivant la
       nature de son argument:
       i) primitive :
       ii) fermeture :
       iii) continuation :
       2) fermeture : on lie les paramtres aux valeurs,
       puis on value (=> goto eval) le corps de la fermeture
       dans le nouvel environnement obtenu par extension de
       l'environnement de la fermeture
       3) continuation : on rtablit la continuation courante
       et l'environnement au moment de la cration de cette
       continuation. La valeur est celle de l'argument, puis
       on retourne  "next". Dans ce cas, CCONT_NEXT est inutile
       puisqu'on abandonne la continuation courante
    */

 apply:
    {
      switch SWIT_TYPE(fun)
	{
	case PRIM0: {
	  if (nargs == 0) {
	  thunk_prim0:
	    val = APPLY0(SWIT_PRIM0(fun));
	    if (val) {
	      ccont = CCONT_NEXT(ccont);
	      goto next;
	    }
	    goto cerror;
	  }
	  prim_err_msg = wrong_arity;
	  exn = exn_arity;
	  exn_val = fun;
	  goto cerror;
	}
	case CURRENT_ENV_PRIM:
	  val = env ? swit_make_env(env) : swit_void_env;
	  ccont = CCONT_NEXT(ccont);
	  goto next;
	case PRIM01: {
	  if (nargs == 1 ) {
	    SWiT_Sym_prim1 prim = SWIT_PRIM1(fun);
	    if (CHECK_(ARG1, prim, args[0], 1)) {
	      val = APPLY1(prim, args[0]);
	      if (val) {
		ccont = CCONT_NEXT(ccont);
		goto next;
	      }
	      goto cerror;
	    } else {
	      prim_err_msg = wrong_type;
	      exn_val = args[0];
	      exn = exn_type;
	      goto cerror;
	    }
	  } else if (nargs == 0) {
	  thunk_prim01:
	    {
	      SWiT_Sym_prim1 prim = SWIT_PRIM1(fun);
	      val = APPLY1(prim, PRIM_VAL(prim));
	      if (val) {
		ccont = CCONT_NEXT(ccont);
		goto next;
	      }
	    }
	  }
	  prim_err_msg = wrong_arity;
	  exn = exn_arity;
	  exn_val = fun;
	  goto cerror;
	}
	case VALUES_PRIM:
	  if (nargs == 0) {
	    prim_err_msg = wrong_arity;
	    exn = exn_arity;
	    obj = NULL; /* provisiore */
	    goto cerror;
	  } else {
	    SWiT *vals;
	    ccont = CCONT_NEXT(ccont);
	    vals = (SWiT *)CCONT_VALS(ccont);
	    if (vals && (SWIT_TYPE(vals[2]) == CALL_WITH_VALUES_PRIM)) {
	      fun = vals[4]; /* the consumer */
	      goto apply;
	    } else if (nargs == 1) {
	      val = args[0];
	      goto next;
	    } else {
	      prim_err_msg = wrong_arity;
	      exn = exn_arity;
	      obj = NULL; /* provisiore */
	      goto cerror;
	    }
	  }
	case PRIM0N: {
	  if (nargs > 0) {
	    SWiT_Sym_prim0n prim = SWIT_PRIM0n(fun);
	    SWiT val0;
	    register int i;
	    for (i = 0; i < nargs-1; i++) {
	      val0 = args[i];
	      if (!CHECK_(ARGS, prim, val0, 1)) {
		prim_err_msg = wrong_type;
		exn = exn_type;
		exn_val = val0;
		goto cerror;
	      }
	    }
	    val0 = args[i];
	    if (!CHECK_(LAST, prim, val0, 1)) {
	      prim_err_msg = wrong_type;
	      exn = exn_type;
	      exn_val = val0;
	      goto cerror;
	    }
	    val = APPLY2(prim, nargs, args);
	  } else {
	  thunk_prim0n:
	    val = APPLY2(SWIT_PRIM0n(fun), 0, NULL);
	  }
	  if (val) {
	    ccont = CCONT_NEXT(ccont);
	    goto next;
	  }
	  goto cerror;
	}
	case CALL_CC_PRIM: {
	  if (nargs == 1) {
	    SWiT arg = args[0];
	    switch SWIT_TYPE(arg)
	      {
	      case PRIM1:
		ccont = CCONT_NEXT(ccont);
		val = APPLY1(SWIT_PRIM1(arg), make_cont(ccont));
		if (val) {
		  goto next;
		}
		goto cerror;
	      case CLOSURE:
		{
		  SWiT_Closure clos = SWIT_CLOSURE(arg);
		  SWiT_Lambda lbd = CLOSURE_LBD(clos);
		  ccont = CCONT_NEXT(ccont);
		  if (LAMBDA_NPARAMS(lbd) == 1) {
		    SWiT *cont = swit_alloc_ptrs(1);
		    cont[0] = make_cont(ccont);
		    env = swit_extend_env(LAMBDA_NAMES(lbd), cont, CLOSURE_ENV(clos));
		    obj = LAMBDA_BODY(lbd);
		    goto eval;
		  } else {
		    prim_err_msg = call_cc_wrong_arg;
		    exn_val = arg;
		    goto cerror;
		  }
		}
	      case CONT:
		{
		  val = make_cont(CCONT_NEXT(ccont));
		  ccont = SWIT_DATA(arg, SWiT_Ccont);
		  goto next;
		}
	      default:
		prim_err_msg = call_cc_wrong_arg;
		exn_val = arg;
		goto cerror;
	      }
	  } else {
	    prim_err_msg = wrong_number;
	    exn = exn_arity;
	    exn_val = fun;
	    goto cerror;
	  }
	}
	case CAR_PRIM: {
	  if (nargs == 1) {
	    SWiT arg = args[0];
	    if SWIT_PAIRP(arg) {
	      val = SWIT_CAR(arg);
	      ccont = CCONT_NEXT(ccont);
	      goto next;
	    }
	    prim_err_msg = not_a_pair;
	    exn_val = arg;
	    exn = exn_type;
	    goto cerror;
	  }
	  prim_err_msg = wrong_arity;
	  exn = exn_arity;
	  exn_val = fun;
	  goto cerror;
	}
	case CDR_PRIM: {
	  if (nargs == 1) {
	    SWiT arg = args[0];
	    if SWIT_PAIRP(arg) {
	      val = SWIT_CDR(arg);
	      ccont = CCONT_NEXT(ccont);
	      goto next;
	    }
	    prim_err_msg = not_a_pair;
	    exn_val = arg;
	    exn = exn_type;
	    goto cerror;
	  }
	  prim_err_msg = wrong_arity;
	  exn = exn_arity;
	  exn_val = fun;
	  goto cerror;
	}
	case PRIM1: {
	  if (nargs == 1) {
	    SWiT_Sym_prim1 prim = SWIT_PRIM1(fun);
	    SWiT arg = args[0];
	    if (CHECK_(ARG1, prim, arg, 1)) {
	      val = APPLY1(prim, arg);
	      if (val) {
		ccont = CCONT_NEXT(ccont);
		goto next;
	      }
	      goto cerror;
	    } else {
	      prim_err_msg = wrong_type;
	      exn = exn_type;
	      exn_val = arg;
	      goto cerror;
	    }
	  } else {
	    prim_err_msg = wrong_arity;
	    exn = exn_arity;
	    exn_val = fun;
	    goto cerror;
	  }
	}
	case RAISE_PRIM:
	  /* check args : A FAIRE */
	  if (nargs == 1) {
	    prim_err_msg = exception;
	    exn = args[0];
	  } else if (nargs == 2) {
	    prim_err_msg = exception;
	    exn = args[0];
	    exn_val = args[1];
	  } else {
	    prim_err_msg = wrong_arity;
	    exn = exn_arity;
	  }
	  goto cerror;

	case PRIM12: {
	  SWiT_Sym_prim2 prim = SWIT_PRIM2(fun);
	  SWiT val1;
	  SWiT val2;
	  if (nargs == 1) {
	    val1 = args[0];
	    val2 = PRIM_VAL(prim);
	  } else if (nargs == 2 ) {
	    val1 = args[0];
	    val2 = args[1];
	  } else {
	    prim_err_msg = wrong_arity;
	    exn = exn_arity;	      
	    exn_val = fun;
	    goto cerror;
	  }
	  if (CHECK_(ARG2, prim, val2, 1)
	      && CHECK_(ARG1, prim, val1, 1)) {
	    val = APPLY2(prim, val1, val2);
	    if (val) {
	      ccont = CCONT_NEXT(ccont);
	      goto next;
	    }
	    goto cerror;
	  } else {
	    prim_err_msg = wrong_type;
	    /* exn_val = val2; / * ou val1 */
	    exn = exn_type;
	  }
	  goto cerror;
	}
	case PRIM1N: {
	  if (nargs > 0) {
	    SWiT_Sym_prim1n prim = SWIT_PRIM1n(fun);
	    SWiT val1 = args[0];
	    if (!CHECK_(ARG1, prim, val1, 1)) {
	      prim_err_msg = wrong_type;
	      exn = exn_type;
	      exn_val = val1;
	      goto cerror;
	    }
	    if (nargs > 1) {
	      register int i;
	      SWiT tmp;
	      for (i = 1; i < nargs-1; i++) {
		tmp = args[i];
		if (!CHECK_(ARGS, prim, tmp, 1)) {
		  prim_err_msg = wrong_type;
		  exn = exn_type;
		  exn_val = tmp;
		  goto cerror;
		}
	      }
	      tmp = args[i];
	      if (!CHECK_(LAST, prim, tmp, 1)) {
		prim_err_msg = wrong_type;
		exn = exn_type;
		exn_val = tmp;
		goto cerror;
	      }
	      val = APPLY3(prim, val1, nargs-1, args+1);
	    } else {
	      val = APPLY3(prim, val1, 0, NULL);
	    }
	    if (val) {
	      ccont = CCONT_NEXT(ccont);
	      goto next;
	    }
	    goto cerror;
	  } else {
	    prim_err_msg = wrong_arity;
	    exn = exn_arity;
	    exn_val = fun;
	  }
	  goto cerror;
	}
	case CALL_WITH_VALUES_PRIM:
	  if (nargs == 2) {
	    fun = args[0]; /* le producteur */
	    if (fun) {
	      switch SWIT_TYPE(fun)
		{
		case CLOSURE: {
		  SWiT_Closure clos;
		  SWiT_Lambda code;
		  clos = SWIT_CLOSURE(fun);
		  code = CLOSURE_LBD(clos);
		  if (LAMBDA_NPARAMS(code) == 0) { /* must be a thunk */
		    args[0] = NULL;
		    args[2] = args[1];
		    obj = LAMBDA_BODY(code);
		    env = swit_extend_env(NULL, NULL, CLOSURE_ENV(clos));
		    goto eval;
		  } else {
		    exn = exn_arity;
		    exn_val = fun;
		    goto cerror;
		  }
		}
		case VALUES_PRIM:
		  fun = args[1]; /* le consommateur */
		  nargs = 0;
		  goto apply;
		case PRIM0:
		  args[0] = APPLY0(SWIT_PRIM0(fun));
		  fun = args[1]; /* le consommateur */
		  nargs = 1;
		  goto apply;
		case PRIM01:
		  {
		    SWiT_Sym_prim1 prim = SWIT_PRIM1(fun);
		    args[0] = APPLY1(prim, PRIM_VAL(prim));
		    fun = args[1]; /* le consommateur */
		    nargs = 1;
		    goto apply;
		  }
		case PRIM0N:
		  args[0] = APPLY2(SWIT_PRIM0n(fun), 0, NULL);
		  fun = args[1]; /* le consommateur */
		  nargs = 1;
		  goto apply;
		default:
		  exn = exn_arity;
		  exn_val = fun;
		  goto cerror;
		}
	    } else {
	      fun = args[2]; /* le consommateur (dplac ici car val est en args[1]) */
	      nargs = 1;
	      args[0] = val;
	      goto apply;
	    }
	  } else {
	    prim_err_msg = wrong_arity;
	    exn = exn_arity;
	    exn_val = fun;
	    goto cerror;
	  }

	case CERROR_PRIM:
	  if (nargs == 2) {
	    SWiT msg = args[0];
	    SWiT expr = args[1];
	    if SWIT_STRINGP(msg) {
	      prim_err_msg = STRING_ELS(msg);
	      CCONT_UNEV(ccont) = swit_make_quote(expr);
	      obj = NULL;
	      goto cerror;
	    }
	    prim_err_msg = wrong_type;
	    exn = exn_type;
	    return NULL;
	  } else {
	    prim_err_msg = wrong_arity;
	    exn = exn_arity;
	    exn_val = fun;
	    goto cerror;
	  }

	case PRIM2:
	  if (nargs == 2 ) {
	    SWiT_Sym_prim2 prim = SWIT_PRIM2(fun);
	    SWiT val1 = args[0];
	    SWiT val2 = args[1];
	    if (CHECK_(ARG1, prim, val1, 1)
		&& CHECK_(ARG2, prim, val2, 1)) {
	      val = APPLY2(prim, val1, val2);
	      if (val) {
		ccont = CCONT_NEXT(ccont);
		goto next;
	      }
	      goto cerror;
	    } else {
	      prim_err_msg = wrong_type;
	      exn_val = val1; /* ou val2 */
	      exn = exn_type;
	    }
	  } else {
	    prim_err_msg = wrong_arity;
	    exn_val = fun;
	    exn = exn_arity;
	  }
	  goto cerror;
	case APPLY_PRIM:
	case PRIM2N: {
	  if (nargs > 1) {
	    SWiT_Sym_prim2n prim = SWIT_PRIM2n(fun);
	    SWiT last;
	    SWiT val1 = args[0];
	    SWiT val2 = args[1];
	    SWiT *vals;
	    if (!CHECK_(ARG1, prim, val1, 1)) {
	      prim_err_msg = wrong_type;
	      exn = exn_type;
	      exn_val = val1;
	      goto cerror;
	    }
	    if (nargs == 2) {
	      if (!CHECK_(ARG2, prim, val2, 1)) {
		prim_err_msg = wrong_type;
		exn = exn_type;
		exn_val = val2;
		goto cerror;
	      }
	      vals = NULL;
	    } else {
	      register int i;
	      if (!CHECK_(ARG2, prim, val2, 1)) {
		prim_err_msg = wrong_type;
		exn = exn_type;
		exn_val = val2;
		goto cerror;
	      }
	      vals = args+2;
	      for (i = 0; i < nargs-3; i++) {
		last = vals[i];
		if (!CHECK_(ARGS, prim, last, 1)) {
		  prim_err_msg = wrong_type;
		  exn = exn_type;
		  exn_val = last;
		  goto cerror;
		}
	      }
	      last = vals[i];
	      if (!CHECK_(LAST, prim, last, 1)) {
		prim_err_msg = wrong_type;
		exn = exn_type;
		exn_val = last;
		goto cerror;
	      }
	    }
	    if (SWIT_TYPE(fun) == APPLY_PRIM) {
	      int num_rands, i;
	      SWiT lst;
	      int len;

	      if (vals == NULL) {
		lst = val2;
	      } else {
		lst = last;
	      }
	      if (lst == swit_null_list) {
		len = 0;
	      } else {
		len = swit_list_length(lst);
		if (len <= 0) {
		  prim_err_msg = improper_list;
		  exn = exn_type;
		  exn_val = lst;
		  goto cerror;
		}
	      }
	      num_rands = nargs - 2 + len;
	      args = swit_alloc_ptrs(num_rands+4);
	      args[0] = val2; /* a1 or l */
	      if (nargs > 2) {
		for (i = 1; i < nargs-2; i++) { /* a2 ... an */
		  args[i] = vals[i-1];
		}
	      }
	      for (i = nargs-2; i < num_rands; i++) {
		args[i] = SWIT_CAR(lst);
		lst = SWIT_CDR(lst);
	      }
	      fun = val1;
	      nargs = num_rands;
	      goto apply;
	    }
	    val = APPLY4(prim, val1, val2, nargs-2, vals);
	    if (val) {
	      ccont = CCONT_NEXT(ccont);
	      goto next;
	    }
	  } else {
	    prim_err_msg = wrong_arity;
	    exn = exn_arity;
	    exn_val = fun;
	  }
	  goto cerror;
	}
	case PRIM3: {
	  if (nargs == 3) {
	    SWiT_Sym_prim3 prim = SWIT_PRIM3(fun);
	    SWiT val1 = args[0];
	    SWiT val2 = args[1];
	    SWiT val3 = args[2];
	    if (CHECK_(ARG1, prim, val1, 1)
		&& CHECK_(ARG2, prim, val2, 1)
		&& CHECK_(ARG3, prim, val3, 1)) {
	      val = APPLY3(prim, val1, val2, val3);
	      if (val) {
		ccont = CCONT_NEXT(ccont);
		goto next;
	      }
	      goto cerror;
	    } else {
	      prim_err_msg = wrong_type;
	      exn_val = val1; /* val1 2 ou 3 */
	      exn = exn_type;
	    }
	  } else {
	    prim_err_msg = wrong_arity;
	    exn = exn_arity;
	    exn_val = fun;
	  }
	  goto cerror;
	}
	case CLOSURE: {
	thunk_closure:
	  {
	    SWiT_Closure clos = SWIT_CLOSURE(fun);
	    SWiT_Lambda code = CLOSURE_LBD(clos);
	    int arity = LAMBDA_NPARAMS(code);
	    ccont = CCONT_NEXT(ccont);
	    obj = LAMBDA_BODY(code);
	    if (arity >> 16) {
	      /* arity >> 16 est le nombre minimal d'arguments */
	      int i;
	      SWiT val_list = swit_null_list;
	      if (!args) {
		exn = exn_arity;
		exn_val = fun;
		goto cerror;
	      }
	      arity = arity & TWO_INT_MASK;
	      if (nargs < arity) goto erreur;
	      if (arity < nargs) {
		for (i = nargs-1; i > arity-1; i--) {
		  val_list = swit_make_pair(args[i], val_list);
		}
		args[arity] = val_list;
	      } else if (arity == 0) { /* => nargs == 0 */
		SWiT *val = swit_alloc_ptrs(1);
		val[0] = swit_null_list;
		env = swit_extend_env(LAMBDA_NAMES(code), val, CLOSURE_ENV(clos));
		goto eval;
	      } else { /* nargs = arity */
		/* on dcale (? ?) */
		args--;
		for (i = 0; i < arity; i++) {
		  args[i] = args[i+1];
		}
		args[i] = val_list;
		env = swit_extend_env(LAMBDA_NAMES(code), args, CLOSURE_ENV(clos));
		goto eval;
	      }
	    } else if (nargs != arity) {
	      goto erreur;
	    }
	    env = swit_extend_env(LAMBDA_NAMES(code), args, CLOSURE_ENV(clos));
	    goto eval;
	  erreur:
	    exn = exn_arity;
	    exn_val = fun;
	    goto cerror;
	  }
	}

	case CONT:
	  if (nargs == 1) {
	    ccont = SWIT_CONT(fun);
	    val = args[0];
	    goto next;
	  }
	  exn = exn_arity;
	  exn_val = fun;
	  goto cerror;

	case ENV: {
	  /* the first argument must be a symbol bound in `fun'
	   * and its value a closure that is called with fun + rest
	   * of arguments, otherwise it is an error.
	   */
	  if (nargs == 0) {
	    exn = exn_arity;
	    prim_err_msg = wrong_arity;
	    exn_val = fun;
	    goto err_env;
	  } else {
	    SWiT sym, clos = NULL;
	    sym = args[0];
	    if SWIT_SYMBOLP(sym) {
	      SWiT_Env e = SWIT_ENV(fun);
	      while (e) {
		SWiT names = ENV_NAMES(e);
		int idx = search(sym, names);
		if (idx != -1) {
		  clos = ENV_FRAME(e)[idx];
		  break;
		}
		e = ENV_NEXT(e);
	      }
	      if (!clos) {
		val = swit_unbound; /* or error (best choice?) */
		ccont = CCONT_NEXT(ccont);
		goto next;
	      }
	    } else {
	      exn = exn_type;
	      prim_err_msg = wrong_type;
	      goto err_env;
	    }
	    if (SWIT_CLOSUREP(clos) &&
		(LAMBDA_NAMES(CLOSURE_LBD(SWIT_CLOSURE(clos)))
		 != swit_null_list)) /* at least 1 arg: `self' */ {
	      args[0] = fun;
	      fun = clos;
	      goto apply;
	    } else if (nargs == 1) {
	      val = clos;
	      ccont = CCONT_NEXT(ccont);
	      goto next;
	    } else {
	      exn = exn_type;
	      prim_err_msg = wrong_type;
	      goto err_env;
	    }
	  }
	err_env:
	  goto cerror;
	}
	default:
	  exn = exn_type;
	  prim_err_msg = not_a_proc;
	  exn_val = fun;
	  goto cerror;
	}
    }
 cerror:
  {
    while (exn_stack) {
      SWiT_Try try = exn_stack->try;
      SWiT exn_lst = TRY_EXN(try);
      SWiT_Expr *handlers = TRY_HDL(try);

      /* search for a handler symbol equal to exn */
      while SWIT_PAIRP(exn_lst) {
	SWiT exn_sym = SWIT_CAR(exn_lst);
	if ((exn_sym == swit_else) || (exn_sym == exn)) {
	  obj = *handlers;
	  env = exn_stack->env;
	  ccont = exn_stack->cont;
	  exn_stack = exn_stack->next;
	  ccont = extend_cont(env, TRY_PROTECT(try), (void *)&exn_val, ccont);
	  goto eval;
	}
	exn_lst = SWIT_CDR(exn_lst);
	handlers++;
      }
      exn_stack = exn_stack->next;
      if (!exn_stack) {
	/* no handler found */
	/* y a-t-il un "protect" ? */
      }
    }
    /* default handler */
    if (tk) {
      return NULL;
    } else {
      int filedes = PORT_DESCR(SWIT_PORT(cur_in_port));
      int issock = issocket(filedes);
      int errinload = (!isatty (filedes)) && !issock && (filedes != -1);
      SWiT fv = make_fv(env);

      /* if transcript on, just write this (to be improved ...) */
      if (transcript_descr != -1) {
	write(transcript_descr, exception_newline, 17);
      }

      /* Announcing message */
      if (errinload) {
	fprintf(stderr, error_at_line, filename, PORT_LINENO(SWIT_PORT(cur_in_port)));
      } else {
	fprintf(stderr, "[CERROR] -- ");
      }

      /* display the error message */
      fprintf(stderr, prim_err_msg, nargs);
      prim_err_msg = unknown_error;

      /* display the object (cause of the error) */
      if (exn_val) {
	fprintf(stderr, " `");
	swit_printl(exn_val, fv);
	fprintf(stderr, "'\n");
	exn_val = NULL;
      }

      /* display the exception name associated to the error */
      fprintf(stderr, uncaught_exception, (char *)SYMBOL_STRING(SWIT_SYM(exn)));

      /* display the expression where the error occured */
      /* --> pas trs joli pour l'instant */
      if ((ccont != done_cont) && (ccont != break_cont)
	  && (SWIT_INST(CCONT_UNEV(ccont)) != PROTECT_CODE)){
	fprintf(stderr, in);
	swit_print_code(CCONT_UNEV(ccont), fv);
      }
      fprintf(stderr,"\n");

      /* try to recover from error (when not loading a file) */
      if (!errinload) {
	SWiT_Expr code;
	SWiT_Ccont sc = ccont;
	SWiT saved_prompt = swit_prompt;
	swit_prompt = swit_break_prompt;
	prompt = 1;
	read_paren = 0;
	ccont = break_cont;
	CCONT_ENV(ccont) = env;
	break_loop:
	env = CCONT_ENV(ccont);
	if (env) {
	  code = read_exp(get_tok(), swit_make_env(env));
	} else {
	  code = read_program(get_tok());
	}
	if ((code != (SWiT_Expr)swit_true)
	    && (code != (SWiT_Expr)swit_false)) {
	  obj = code;
	  goto eval;
	}
	ccont = sc;
	swit_prompt = saved_prompt;
	if (obj && (code == (SWiT_Expr)swit_true)) {
	  goto eval;
	}
      }
      return NULL;
    }
  }
}
