/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/
/*--------------------------------------------------------------------
 * This file initialize miscellaneous swit global variables
 * and constants and call all other initialization procedures
 * defined elsewhere.
 *--------------------------------------------------------------------
 */

#include "swit.h"

#ifdef ENGLISH
#include "messages_english.h"
#endif
#ifdef FRENCH
#include "messages_french.h"
#endif

extern void swit_init_read (void);
extern void swit_init_boolean (void);
extern void swit_init_char (void);
extern void swit_init_equiv (void);
extern void swit_init_pair (void);
extern void swit_init_symbol (void);
extern void swit_init_string (void);
extern void swit_init_vector (void);
extern void swit_init_quad (void);
extern void swit_init_model (void);
extern void swit_init_io (void);
extern void swit_init_print (void);
extern void swit_init_control (void);
extern void swit_init_number (void);
extern void swit_init_misc (void);
extern void swit_init_top (void);
extern void swit_init_queue (void);
extern void swit_init_socket (void);
extern void swit_init_type (void);
extern void swit_init_code (void);
extern void swit_init_disassemble (void);

extern void swit_load_init (char *filename);
extern SWiT cur_in_port;
extern int file_exists (char *filename);
extern SWiT swit_load (char *file);
extern jmp_buf *error_Ptr;
extern void reset_globals (void);
extern int interrupt;
extern int have_tk;

SWiT default_print_depth;
SWiT default_print_span;
char *implementation_version = "0.97.3";
char *default_lib_path = "/usr/local/lib/swit";
/* implementation_date : defined in messages.h */

static jmp_buf init_error; /* just in case initialisation script fails */
static char *break_prompt = "break >> ";

/*
 * Global symbols :
 */
SWiT swit_eof;
SWiT swit_zero;
SWiT swit_one;
SWiT swit_unbound;
SWiT swit_null_list;
SWiT swit_version;
SWiT swit_prompt;
SWiT swit_type_flag;

/* exception symbols */
SWiT exn;
SWiT exn_val;
SWiT exn_type;
SWiT exn_range;
SWiT exn_arity;
SWiT exn_unbound;
SWiT exn_math;
SWiT exn_io;
SWiT exn_read;
SWiT exn_unknown;
SWiT exn_cont;
SWiT exn_env;
SWiT exn_size;


/* 'User' global variables */
SWiT swit_time;
SWiT swit_load_path;
SWiT swit_verbose;
SWiT swit_print_depth;
SWiT swit_print_span;
SWiT swit_print_cycle;
SWiT swit_void_env;
SWiT swit_break_prompt;
SWiT swit_nil;
SWiT swit_error_object;
SWiT swit_magic;

/* Banner */
void
swit_banner (void)
{
    fprintf(stdout, banner_msg, implementation_version, implementation_date);
  if (have_tk) {
    fprintf(stdout, browse_msg);
  }
}

/* global tables to be filled inside each module */
#define TABLE_DIM 281
void *read_table[TABLE_DIM];
void *equal_table[TABLE_DIM];

#define current_ip SWIT_PORT(cur_in_port)
#define read_jbuf PORT_JBUF(current_ip)
static void *
read_error (void)
{
  exn = exn_read;
  prim_err_msg = unexpected_token;
  longjmp(*read_jbuf, 2);
  return NULL;
}

static int
ret_zero (SWiT obj1, SWiT obj2)
{
  return 0;
}

static void
init_tables ()
{
  int i;
  for (i = 0; i < TABLE_DIM; i++) {
    read_table[i] = read_error;
    equal_table[i] = ret_zero;
  }
}

/*
 * Some constants are defined only with their type
 * (i.e. they are unique in their type)
 */
static SWiT
make_constant (int typ)
{
  SWiT obj = swit_alloc_object();
  SWIT_TYPE(obj) = typ;
  obj->data = (SWiT)0xffffffff;
  return (obj);
}

extern SWiT_Ccont init_cont (int inst);

/* Warning: from code.h */
#define DONE 92
#define BREAK 94
/* exported to switEval.c, switIo.c, switMisc.c, switRead.c */
SWiT_Ccont done_cont;
SWiT_Ccont break_cont;

static void
swit_init_constants (void)
{
  done_cont = init_cont(DONE);
  break_cont = init_cont(BREAK);
  swit_zero = swit_make_fixnum(0);
  swit_one = swit_make_fixnum(1);
  swit_null_string = swit_make_constant_string(NULL, 0);
  swit_null_vect = swit_make_vector(0, NULL);
  swit_version = swit_make_constant_string(implementation_version, strlen(implementation_version));
  swit_break_prompt = swit_make_symbol("*break-prompt*", 14, swit_make_string(break_prompt, strlen(break_prompt)));
  default_print_depth = swit_make_fixnum(50);
  default_print_span = swit_make_fixnum(1000);

  /*
   * GLOBAL SYMBOLS  :
   */
  swit_nil = swit_make_symbol("nil", 3, swit_null_list);
  swit_error_object = swit_make_symbol("#<error>", 8, swit_unbound);
  swit_time = swit_make_symbol("*time*", 6, swit_zero);
  swit_verbose = swit_make_symbol("*verbose*", 9, swit_false);
  swit_prompt = swit_make_symbol("*prompt*", 8, swit_make_string("swit> ", 2));
  swit_load_path = swit_make_symbol("*load-path*", 11, swit_null_string);
  swit_print_depth = swit_make_symbol("*print-depth*", 13, default_print_depth);
  swit_print_span = swit_make_symbol("*print-length*", 14, default_print_span);
  swit_print_cycle = swit_make_symbol("*print-cycle*", 14, swit_false);
  swit_type_flag = swit_make_symbol("*type*", 6, swit_true);
  swit_magic = swit_make_symbol(":*%*%%?", 7, swit_unbound);


  /* build-in exceptions */
  exn_type = swit_make_symbol("error:type", 10, swit_unbound);
  exn_range = swit_make_symbol("error:range", 11, swit_unbound);
  exn_arity = swit_make_symbol("error:arity", 11, swit_unbound);
  exn_unbound = swit_make_symbol("error:unbound", 13, swit_unbound);
  exn_math = swit_make_symbol("error:math", 10, swit_unbound);
  exn_size = swit_make_symbol("error:size", 10, swit_unbound);
  exn_env = swit_make_symbol("error:env", 9, swit_unbound);
  exn_io = swit_make_symbol("error:io", 8, swit_unbound);
  exn_read = swit_make_symbol("error:read", 10, swit_unbound);
  exn_cont = swit_make_symbol("error:continuation", 18, swit_unbound);
  exn_unknown = swit_make_symbol("error:unknown", 13, swit_unbound);
  exn_val = swit_make_symbol("Unknow error", 12, swit_unbound);
}


/*
 *        These are needed by gmp :
 */
static void *
mp_realloc(void * obj, size_t old, size_t new)
{
  return (swit_realloc(obj, new));
}

static void
mp_free(void *obj, size_t s)
{
  return (swit_free(obj));
}

void
swit_init (void)
{
  interrupt = 0;

  /* memory initialisation for gmp */
  mp_set_memory_functions(swit_malloc, mp_realloc, mp_free);

  /* DON'T CHANGE INITIALISATION ORDER */
  swit_null_list = make_constant(NIL);
  swit_eof = make_constant(EOF_TYPE);
  swit_unbound =  make_constant(UNBOUND);
  swit_void_env =  swit_make_env((SWiT_Env)NULL);

  init_tables ();

  /* init symbols first because of  primitives_table and global_env */
  swit_init_symbol ();

  /* boolean then, because #t and #f are defined there */
  swit_init_boolean ();

  /* constants that are used everywhere */
  swit_init_constants ();

  /* the order of the rest should not be important */
  swit_init_char ();
  swit_init_string ();
  swit_init_equiv ();
  swit_init_pair ();
  swit_init_number ();
  swit_init_vector ();
  swit_init_control ();
  swit_init_io ();
  swit_init_read ();
  swit_init_code ();
  swit_init_print ();
  swit_init_type ();
  swit_init_misc ();
  swit_init_queue ();
  swit_init_socket ();
  swit_init_quad ();
  swit_init_model ();
  swit_init_disassemble();

  /* last one */
  swit_init_top ();
}

#define current_ip SWIT_PORT(cur_in_port)
#define lineno PORT_LINENO(current_ip)

void
swit_load_objs (void)
{
  char *file;
  error_Ptr = &init_error;
  if ((file=(char *)setjmp(init_error))) {
    fprintf(stderr, load_init, lineno, file);
  } else {
    char filename[256];
    char *lib_path = getenv("SWIT_LIBRARY_PATH");
    if (!lib_path || !*lib_path) {
      if (file_exists(default_lib_path) != -1) {
	lib_path = default_lib_path;
      } else if (file_exists("lib/compile.obj") != -1) {
	lib_path = "lib";
      } else {
	fprintf(stderr, lib_path_error);
	exit(1);
      }
    }
    /* Loading initialisation files */
    swit_load_init("/compile.obj");
    swit_load_init("/map.obj");
    swit_load_init("/disassemble.obj");
    swit_load_init("/sig.obj");
    swit_load_init("/pptype.obj");
    swit_load_init("/unif.obj");
    swit_load_init("/type.obj");
    swit_load_init("/doc.obj");
    strncpy(filename, getenv("HOME"), 240);
    strcat(filename, "/.switinitrc");
    if (file_exists(filename) == -1) {
      strncpy(filename, lib_path, 240);
      strcat(filename, "/switinitrc");
    }
    swit_load(filename);
  }
  reset_globals();
}

void
start_swit (int argc, char **argv)
{
  int cnt = 1;
  swit_init();
  swit_load_objs();
  fprintf(stdout, no_tk);
  /* If (scheme) files was given on the command line: load them */
  while (cnt != argc) {
    char *file = argv[cnt++];
    error_Ptr = &init_error;
    if (file_exists(file) == 0) {
      if (setjmp(init_error)) {
	/* skip */
	fprintf(stderr, load_err, file);
      } else {
	swit_load(file);
      }
    } else {
      fprintf(stderr, file_not_found, file);
    }
    reset_globals();
  }
  swit_repl();
}
