/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*==================================================================
 *                     6.3 PAIRS AND LISTS
 *==================================================================
 * Standard R4RS procedures :
 * ^^^^^^^^^^^^^^^^^^^^^^^^
 *
 *   pair?                    OBJ               -> BOOLEAN
 *   cons                     OBJ X OBJ         -> PAIR
 *   car,cdr                  PAIR              -> OBJ
 *   set-car!                 PAIR X OBJ        -> OBJ
 *   set-cdr!                 PAIR X OBJ        -> OBJ
 *   c[ad]+r                  LIST              -> OBJ
 *   null?                    OBJ               -> BOOLEAN
 *   list?                    OBJ               -> BOOLEAN
 *   list                     OBJ X ...         -> LIST
 *   length                   LIST              -> NUMBER
 *   append                   LIST X ...        -> LIST
 *   reverse                  LIST              -> LIST
 *   list-tail                LIST X NUMBER     -> LIST
 *   list-ref                 LIST X NUMBER     -> OBJ
 *   memq,memv,member         OBJ X LIST        -> LIST or FALSE
 *   assq,assv,assoc          OBJ X ALIST       -> LIST or FALSE
 *
 * Extensions :
 * ^^^^^^^^^^
 *   append!                  LIST X LIST       -> LIST
 *   reverse!                 LIST              -> LIST
 *   list-insert!             LIST X LIST       -> LIST
 *   last-pair                PAIR              -> PAIR
 *   cons*                    OBJ X OBJ X ...   -> LIST
 *   acons                    OBJ X 'B X 'C     -> PAIR
 *   copy-list                LIST              -> LIST
 *   copy-tree                LIST              -> LIST
 *   copy-graph               LIST              -> LIST
 *   remq                     OBJ X LIST        -> LIST
 *   remq!                    OBJ X LIST        -> LIST
 *   remove                   OBJ X LIST        -> LIST
 *   remove!                  OBJ X LIST        -> LIST
 *   subst                    OBJ X OBJ X LIST  -> LIST
 *   sublis                   ALIST X LIST      -> LIST
 *==================================================================
 */

#include "swit.h"

/* exported to Equiv and Print : */
SWiT EX;
SWiT M00;
SWiT M01;
SWiT M10;
SWiT M11;

/*==================================================================
 * forward declarations of procedures defined in this file
 *==================================================================
 */
static SWiT pair_p_prim (SWiT arg);
static SWiT car_prim (SWiT arg);
static SWiT cdr_prim (SWiT arg);
static SWiT set_car_prim (SWiT arg1, SWiT arg2);
static SWiT set_cdr_prim (SWiT arg1, SWiT arg2);
static SWiT caar_prim (SWiT arg);
static SWiT cadr_prim (SWiT arg);
static SWiT cdar_prim (SWiT arg);
static SWiT cddr_prim (SWiT arg);
static SWiT caaar_prim (SWiT arg);
static SWiT caadr_prim (SWiT arg);
static SWiT cadar_prim (SWiT arg);
static SWiT cdaar_prim (SWiT arg);
static SWiT cdadr_prim (SWiT arg);
static SWiT cddar_prim (SWiT arg);
static SWiT caddr_prim (SWiT arg);
static SWiT cdddr_prim (SWiT arg);
static SWiT caaaar_prim (SWiT arg);
static SWiT caaadr_prim (SWiT arg);
static SWiT caadar_prim (SWiT arg);
static SWiT cadaar_prim (SWiT arg);
static SWiT cadadr_prim (SWiT arg);
static SWiT caddar_prim (SWiT arg);
static SWiT caaddr_prim (SWiT arg);
static SWiT cadddr_prim (SWiT arg);
static SWiT cdaaar_prim (SWiT arg);
static SWiT cdaadr_prim (SWiT arg);
static SWiT cdadar_prim (SWiT arg);
static SWiT cddaar_prim (SWiT arg);
static SWiT cddadr_prim (SWiT arg);
static SWiT cdddar_prim (SWiT arg);
static SWiT cdaddr_prim (SWiT arg);
static SWiT cddddr_prim (SWiT arg);
static SWiT null_p_prim (SWiT arg);
static SWiT list_p_prim (SWiT arg);
static SWiT memv_prim (SWiT arg1, SWiT arg2);
static SWiT memq_prim (SWiT arg1, SWiT arg2);
static SWiT member_prim (SWiT arg1, SWiT arg2);
static SWiT assv_prim (SWiT arg1, SWiT arg2);
static SWiT assq_prim (SWiT arg1, SWiT arg2);
static SWiT assoc_prim (SWiT arg1, SWiT arg2);
static SWiT reverse_prim (SWiT arg);
static SWiT list_tail_prim (SWiT arg1, SWiT arg2);
static SWiT list_ref_prim (SWiT arg1, SWiT arg2);
static SWiT length_prim (SWiT arg);
static SWiT list_prim (int argc, SWiT *argv);
static SWiT append_prim (int argc, SWiT *argv);
static SWiT nconc_prim (int argc, SWiT *argv);
static SWiT nreverse_prim (SWiT arg);
static SWiT liststar_prim (SWiT arg1, SWiT arg2, int argc, SWiT *argv);
static SWiT acons_prim (SWiT arg1, SWiT arg2, SWiT arg3);
static SWiT copy_list_prim (SWiT arg);
static SWiT copy_graph_prim (SWiT arg);
static SWiT remq_prim (SWiT obj, SWiT list);
static SWiT remove_prim (SWiT obj, SWiT list);
static SWiT nremq_prim (SWiT obj, SWiT list);
static SWiT nremove_prim (SWiT obj, SWiT list);
static SWiT create_mark (int m);

/*==================================================================
 *              PRINTING LISTS AND GRAPHS
 *==================================================================
 */

extern void mark (SWiT expr);
extern SWiT peel (SWiT oldcel);
extern SWiT EX;
extern SWiT M00;
extern SWiT M01;
extern SWiT M10;
extern SWiT M11;
extern SWiT swit_print_cycle;

/* exported : to switTop.c (for reinitialization), to switPrint.c */
int print_span;
int print_depth;

static void print_graph (SWiT expr, int esc, int tcl, SWiT fv, int depth);


#define print_open_paren(tcl) char2buff(tcl?'{':'(')
#define print_close_paren(tcl) char2buff(tcl?'}':')')

static void
print_atom (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  if (obj == swit_quote) {
    str2buff("quote");
  } else if (obj == swit_unquote) {
    str2buff("unquote");
  } else if (obj == swit_quasiquote) {
    str2buff("quasiquote");
  } else if (obj == swit_unquote_splicing) {
    str2buff("unquote-splicing");
  } else {
    print_to_buffer(obj, esc, tcl, fv, depth);
  }
}


#define M(cell) (SWIT_CAR(cell))
#define A(cell) (SWIT_CAR(SWIT_CDR(cell)))
#define D(cell) (SWIT_CDR(SWIT_CDR(cell)))

static void
print_graph_list (SWiT expr, int esc, int tcl, SWiT fv, int depth)
{
  SWiT mrk = M(expr);
  SWiT car = A(expr);
  SWiT cdr = D(expr);
  if ((mrk == M00) || (mrk == M01)) {
    print_graph(car, esc, tcl, fv, depth);
  } else {  /* M10 M11 */
    if (SWIT_PAIRP(car)) {
      char2buff('#');
    } else {
      print_atom(car, esc, tcl, fv, depth);
    }
  }
  if ((mrk == M01) || (mrk == M11)) {
    if (SWIT_PAIRP(cdr)) {
      str2buff(" . #");
      print_close_paren(tcl);
    } else if (cdr == swit_null_list) {
      print_close_paren(tcl);
    } else {
      str2buff(" . ");
      print_atom(cdr, esc, tcl, fv, depth);
      print_close_paren(tcl);
    }
  } else { /* M00 M10 */
    if (SWIT_PAIRP(cdr)) {
      space2buff();
      print_graph_list(cdr, esc, tcl, fv, depth);
    } else if (cdr == swit_null_list) {
      print_close_paren(tcl);
    } else {
      str2buff(" . ");
      print_atom(cdr, esc, tcl, fv, depth);
      print_close_paren(tcl);
    }
  }
}

static void
print_graph (SWiT expr, int esc, int tcl, SWiT fv, int depth)
{
  if (SWIT_PAIRP(expr)) {
    print_open_paren(tcl);
    print_graph_list(expr, esc, tcl, fv, depth);
  } else {
    print_atom(expr, esc, tcl, fv, depth);
  }
}


void print_pair_std (SWiT, int, int, int, SWiT, int);

void
print_pair (SWiT expr, int esc, int tcl, SWiT fv, int depth)
{
  if (SYMBOL_VAL(SWIT_SYM(swit_print_cycle)) == swit_true) {
    mark(expr);
    print_graph(expr, esc, tcl, fv, depth);
    peel(expr);
  } else {
    print_pair_std(expr, 0, esc, tcl, fv, 0);
  }
}

static void
rec_print_list (SWiT obj, int pflag, int esc, int tcl, SWiT fv, int length, int depth);

/* exported to switDisassemble.c */
void
print_pair_std (SWiT expr, int pflag, int esc, int tcl, SWiT fv, int depth)
{
  if SWIT_PAIRP(expr) {
    SWiT car = SWIT_CAR(expr);
    SWiT cdr = SWIT_CDR(expr);
    if SWIT_PAIRP(cdr) {
      SWiT cadr = SWIT_CAR(cdr);
      if (pflag) {
	if (car == swit_quote) {
	  char2buff('\'');
	  print_pair_std (cadr, 0, esc, tcl, fv, depth);
	} else if (car == swit_unquote) {
	  char2buff(',');
	  print_pair_std (cadr, pflag, esc, tcl, fv, depth);
	} else if (car == swit_quasiquote) {
	  char2buff('`');
	  print_pair_std (cadr, pflag, esc, tcl, fv, depth);
	} else if (car == swit_unquote_splicing) {
	  str2buff(",@");
	  print_pair_std (cadr, pflag, esc, tcl, fv, depth);
	} else {
	  print_open_paren(tcl);
	  rec_print_list(expr, pflag, esc, tcl, fv, 0, 1+depth);
	}
      } else {
	print_open_paren(tcl);
	rec_print_list(expr, pflag, esc, tcl, fv, 0, 1+depth);
      }
    } else {
      print_open_paren(tcl);
      rec_print_list(expr, pflag, esc, tcl, fv, 0, 1+depth);
    }
  } else {
    print_atom(expr, esc, tcl, fv, depth);
  }
}

static void
rec_print_list (SWiT obj, int pflag, int esc, int tcl, SWiT fv, int length, int depth)
{
  if (length > print_span) {
    str2buff("....");
    print_close_paren(tcl);
  } else if (depth > print_depth) {
    str2buff("..");
  } else {
    SWiT d = SWIT_CDR(obj);
    print_pair_std (SWIT_CAR(obj), pflag, esc, tcl, fv, depth);
    if (SWIT_PAIRP(d)) {
      space2buff();
      rec_print_list(d, pflag, esc, tcl, fv, 1+length, depth);
    } else if (d == swit_null_list) {
      print_close_paren(tcl);
    } else {
      str2buff(" . ");
      print_atom(d, esc, tcl, fv, depth);
      print_close_paren(tcl);
    }
  }
}


static void
print_mark (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  switch (SWIT_FIXNUM(obj)) {
  case 0:
    str2buff("#<EX>");
    break;
  case 1:
    str2buff("#<M00>");
    break;
  case 2:
    str2buff("#<M01>");
    break;
  case 3:
    str2buff("#<M10>");
    break;
  default:
    str2buff("#<M11>");
  }
}

static void
print_nil (SWiT obj, int esc, int tcl, SWiT fv, int depth)
{
  if (tcl) {
    str2buff("{}");
  } else {
    str2buff("()");
  }
}


/*==================================================================
 *            CONSTRUCTION (see definition in swit.h)
 *==================================================================
 */

SWiT
swit_make_pair (SWiT car, SWiT cdr)
{
  SWiT cell = swit_alloc_object();
  SET_CAR(cell, car);
  SET_CDR(cell, cdr);
  return (cell);
}

/*==================================================================
 *  SYNTAX : <datum list> == ( <datum>* ) | ( <datum>+ . <datum> )
 *==================================================================
 */
extern SWiT cur_in_port;
#define current_ip SWIT_PORT(cur_in_port)
#define read_jbuf PORT_JBUF(current_ip)
static SWiT
read_pair (void)
{
  int token = get_tok();
  SWiT lst, last;
  if ((token == ')') || (token == ']')) {
    return swit_null_list;
  }
  if (token == '.') {
    prim_err_msg = unexpected_token;
    exn = exn_read;
    longjmp(*read_jbuf, 2);
  }
  lst = last = swit_make_pair(swit_null_list, swit_null_list);
  while ((token != ')') && (token != ']') && (token != '.')) {
    SWiT tmp = last, obj = rd_datum(token);
    last = swit_make_pair(obj, swit_null_list);
    SET_CDR(tmp, last);
    token = get_tok();
  }
  if (token == '.') {
    SET_CDR(last, rd_datum(get_tok()));
    if ((get_tok() != ')') && (get_tok() != ']')) {
      prim_err_msg = unexpected_token;
      exn = exn_read;
      longjmp(*read_jbuf, 2);
    }
  }
  return SWIT_CDR(lst);
}


/*================================================
 *                 PRIMITIVES
 *================================================
 */

static SWiT
pair_p_prim (SWiT arg)
{
  return (SWIT_PAIRP(arg) ? swit_true : swit_false);
}

static SWiT
car_prim (SWiT arg)
{
  return(SWIT_CAR(arg));
}


static SWiT
cdr_prim (SWiT arg)
{
  return(SWIT_CDR(arg));
}

static SWiT
set_car_prim (SWiT arg1, SWiT arg2)
{
  SET_CAR(arg1, arg2);
  return (arg2);
}

static SWiT
set_cdr_prim (SWiT arg1, SWiT arg2)
{
  SET_CDR(arg1, arg2);
  return (arg2);
}


/* For 'c....r' other than 'car' and 'cdr', the evaluator
 *  does not check the argument type : thus, this checking
 *  is done here.
 */

#define GEN_CXXR(name,xx,X) \
static SWiT \
name ## _prim (SWiT arg)\
{\
  SWiT ad = SWIT_C ## X ## R(arg); \
  if SWIT_PAIRP(ad) \
      return (c ## xx ## r_prim (ad));\
  prim_err_msg = not_a_pair;\
  exn = exn_type;\
  exn_val = ad; \
  return NULL;\
}

GEN_CXXR(caar,a,A)
GEN_CXXR(cadr,a,D)
GEN_CXXR(cdar,d,A)
GEN_CXXR(cddr,d,D)
GEN_CXXR(caaar,aa,A)
GEN_CXXR(caadr,aa,D)
GEN_CXXR(cadar,ad,A)
GEN_CXXR(caddr,ad,D)
GEN_CXXR(cdaar,da,A)
GEN_CXXR(cdadr,da,D)
GEN_CXXR(cddar,dd,A)
GEN_CXXR(cdddr,dd,D)
GEN_CXXR(caaaar,aaa,A)
GEN_CXXR(caaadr,aaa,D)
GEN_CXXR(caadar,aad,A)
GEN_CXXR(caaddr,aad,D)
GEN_CXXR(cadaar,ada,A)
GEN_CXXR(cadadr,ada,D)
GEN_CXXR(caddar,add,A)
GEN_CXXR(cadddr,add,D)
GEN_CXXR(cdaaar,daa,A)
GEN_CXXR(cdaadr,daa,D)
GEN_CXXR(cdadar,dad,A)
GEN_CXXR(cdaddr,dad,D)
GEN_CXXR(cddaar,dda,A)
GEN_CXXR(cddadr,dda,D)
GEN_CXXR(cdddar,ddd,A)
GEN_CXXR(cddddr,ddd,D)


static SWiT
null_p_prim (SWiT arg)
{
  return ((arg == swit_null_list) ? swit_true : swit_false);
}

static SWiT
list_p_prim (SWiT arg)
{
  SWiT fast, slow;

  fast = slow = arg;
  do {
    if (SWIT_NULLP (fast)) {
      return (swit_true);
    }
    if (! SWIT_PAIRP (fast)) {
      return (swit_false);
    }
    fast = SWIT_CDR (fast);
    if (SWIT_NULLP (fast)) {
      return (swit_true);
    }
    if (! SWIT_PAIRP (fast)) {
      return (swit_false);
    }
    fast = SWIT_CDR (fast);
    slow = SWIT_CDR (slow);
  }
  while (fast != slow);
  return (swit_false);
}



#ifdef STRICT
#define GEN_MEM(name, comp) \
static SWiT \
name ## _prim (SWiT arg1, SWiT arg2) \
{ \
    SWiT list; \
    list = arg2; \
    while (SWIT_PAIRP (list)) { \
      if (comp (arg1, SWIT_CAR (list))) { \
	 return (list); \
      } \
      list = SWIT_CDR (list); \
    }\
    if (list != swit_null_list) { \
      prim_err_msg = proper_list; \
      exn = exn_type;\
      exn_val = arg2;\
      return NULL; \
    } else { \
      return (swit_false); \
    }\
}
#else
#define GEN_MEM(name, comp) \
static SWiT \
name ## _prim (SWiT arg1, SWiT arg2) \
{ \
  SWiT list; \
  list = arg2; \
  while (SWIT_PAIRP (list)) { \
      if (comp (arg1, SWIT_CAR (list))) { \
          return (list); \
      } \
      list = SWIT_CDR (list); \
    }\
  return (swit_false); \
}
#endif

GEN_MEM(memv, swit_eqv)
GEN_MEM(memq, swit_eq)
GEN_MEM(member, swit_equal)


#define GEN_ASS(name, comp) \
static SWiT \
name ## _prim (SWiT arg1, SWiT arg2) \
{ \
  SWiT pair, list; \
  list = arg2; \
  while (SWIT_PAIRP (list)) { \
      pair = SWIT_CAR (list); \
      if (!SWIT_PAIRP(pair)) {\
         prim_err_msg = list_of_pairs;\
         exn = exn_type;\
         exn_val = arg2;\
         return NULL;\
      }\
      if (comp (arg1, SWIT_CAR (pair))) { \
          return (pair); \
      } \
      list = SWIT_CDR (list); \
    } \
  return (swit_false); \
}

GEN_ASS(assv, swit_eqv)
GEN_ASS(assq, swit_eq)
GEN_ASS(assoc, swit_equal)

static SWiT
reverse_prim (SWiT arg)
{
  SWiT last = swit_null_list;
  SWiT lst = arg;
  while (SWIT_PAIRP (lst)) {
    last = swit_make_pair (SWIT_CAR (lst), last);
    lst = SWIT_CDR (lst);
  }
#ifdef STRICT
  if (lst != swit_null_list) {
    prim_err_msg = proper_list;
    exn = exn_type;
    exn_val = arg;
    return NULL;
  }
#endif
  return (last);
}

static SWiT
list_tail_prim (SWiT arg1, SWiT arg2)
{
  SWiT lst = arg1;
  int k = SWIT_FIXNUM(arg2);
  register i;
  if (k < 0) {
    prim_err_msg = index_out_of_range;
    exn = exn_range;
    exn_val = arg2;
    return NULL;
  }
  for (i = 0; i < k; i++) {
    if (SWIT_PAIRP(lst)) {
      lst = SWIT_CDR (lst);
    } else {
      prim_err_msg = index_too_large;
      exn = exn_range;
      exn_val = arg2;
      return NULL;
    }
  }
  return (lst);
}

SWiT
swit_last_pair (SWiT arg)
{
  SWiT lst;

  do {
    lst = arg;
    arg = SWIT_CDR (arg);
  }
  while (SWIT_PAIRP(arg));
  return (lst);
}

static SWiT
list_ref_prim (SWiT arg1, SWiT arg2)
{
  SWiT lst = arg1;
  int k = SWIT_FIXNUM(arg2);
  register i;
  if (k < 0) {
    prim_err_msg = index_out_of_range;
    exn = exn_range;
    exn_val = arg2;
    return NULL;
  }
  for (i = 0; i < k; i++) {
    if SWIT_PAIRP(lst) {
      lst = SWIT_CDR(lst);
    } else {
      prim_err_msg = index_too_large;
      exn = exn_range;
      exn_val = arg2;
      return NULL;
    }
  }
  if SWIT_PAIRP(lst) {
    return SWIT_CAR(lst);
  } else {
    prim_err_msg = index_too_large;
    exn = exn_range;
    exn_val = arg2;
    return NULL;
  }
}

int
swit_list_length (SWiT arg)
{
  SWiT fast, slow;
  register len = 0;
  fast = slow = arg;
  do {
    if (! SWIT_PAIRP (fast)) {
      return (SWIT_NULLP(fast) ? len : -len);
    }
    fast = SWIT_CDR (fast);
    len++;
    if (! SWIT_PAIRP (fast)) {
      return (SWIT_NULLP(fast) ? len : -len);
    }
    fast = SWIT_CDR (fast);
    len++;
    slow = SWIT_CDR (slow);
  }
  while (fast != slow);
  return (0); /* there is a cycle */
}

static SWiT
length_prim (SWiT arg)
{
  if (SWIT_NULLP(arg)) {
    return swit_zero;
  } else {
    int len = swit_list_length (arg);
    if (len) {
      return (swit_make_fixnum(len));
    } else {
      return swit_false;
    }
  }
}

static SWiT
list_prim (int argc , SWiT *args)
{
  SWiT ret = swit_null_list;
  register i;
  for (i = argc-1; i > -1; i--) {
    ret = swit_make_pair(args[i],ret);
  }
  return (ret);
}

SWiT
swit_append2 (SWiT lst1, SWiT lst2)
{
  if (SWIT_PAIRP(lst1)) {
    SWiT last, first, lst = lst1;
    first = last = swit_make_pair(SWIT_CAR(lst), swit_null_list);
    SET_CAR(first, (SWiT)(((unsigned int)SWIT_CAR(lst)) | ((unsigned int)(lst->data) & 0x7)));
    lst = SWIT_CDR(lst);
    while (SWIT_PAIRP(lst)) {
      SWiT pair = swit_make_pair(SWIT_CAR(lst), swit_null_list);
      SET_CAR(pair, (SWiT)(((unsigned int)SWIT_CAR(lst)) | ((unsigned int)(lst->data) & 0x7)));
      SET_CDR(last, pair);
      last = pair;
      lst = SWIT_CDR(lst);
    }
#ifdef STRICT
    if (! SWIT_NULLP(lst)) {
      prim_err_msg = proper_list;
      exn = exn_type;
      exn_val = lst1;
      return NULL;
    }
#endif
    SET_CDR(last, lst2);
    return(first);
  } else if (SWIT_NULLP(lst1)) {
    return (lst2);
  } else {
    exn = exn_type;
    exn_val = lst1;
    return NULL;
  }
}


static SWiT
append_prim (int argc, SWiT *args)
{
  switch (argc) {
  case 0:
    return (swit_null_list);
  case 1:
    return (args[0]);
  default:
    {
      SWiT res = args[argc-1];
      register i;
      for (i = argc-2; i > -1; i--) {
	SWiT arg = args[i];
	res = swit_append2(arg, res);
      }
      return (res);
    }
  }
}


/*=====================================================
 *              E X T E N S I O N S
 *=====================================================
 */

#define GEN_RASS(name, comp) \
static SWiT \
name ## _prim (SWiT arg1, SWiT arg2) \
{ \
  SWiT pair, list; \
  list = arg2; \
  while SWIT_PAIRP (list) { \
      pair = SWIT_CAR (list); \
      if (!SWIT_PAIRP(pair)) {\
         prim_err_msg = list_of_pairs;\
         exn = exn_type;\
	 exn_val = arg2;\
         return NULL;\
      }\
      if (comp (arg1, SWIT_CDR (pair))) { \
          return (pair); \
      } \
      list = SWIT_CDR (list); \
    } \
  return (swit_false); \
}

GEN_RASS(rassv, swit_eqv)
GEN_RASS(rassq, swit_eq)
GEN_RASS(rassoc, swit_equal)

static SWiT
liststar_prim (SWiT arg1, SWiT arg2, int argc , SWiT *args)
{
  if (argc == 0) {
    return (swit_make_pair(arg1, arg2));
  } else {
    SWiT ret = args[argc-1];
    register i;
    for(i = argc-2; i > -1; i--) {
      ret = swit_make_pair(args[i], ret);
    }
    return (swit_make_pair(arg1, swit_make_pair(arg2, ret)));
  }
}

static SWiT
nconc2 (SWiT lst1, SWiT lst2)
{
  if (lst1 == swit_null_list) {
    return (lst2);
  }
  if (lst2 == swit_null_list) {
    return (lst1);
  } else {
    SWiT pair = swit_last_pair(lst1);
#ifdef STRICT
    if (SWIT_CDR(pair) != swit_null_list) {
      prim_err_msg = proper_list1;
      exn = exn_type;
      exn_val = lst1;
      return NULL;
    }
#endif
    SET_CDR(pair, lst2);
    return(lst1);
  }
}

static SWiT
nconc_prim (int argc, SWiT *args)
{
  switch (argc) {
  case 0:
    return swit_null_list;
  case 1:
    return (args[0]);
  default: {
    SWiT res = args[argc-1];
    register i;
    for (i = argc-2; i > -1; i--) {
      SWiT arg = args[i];
      res = nconc2(arg, res);
    }
    return res;
  }
  }
}

static SWiT
nreverse_prim (SWiT arg)
{
  if (SWIT_PAIRP(arg)) {
    SWiT first = arg;
    SWiT next = SWIT_CDR(arg);
    SET_CDR(arg, swit_null_list);
    while (SWIT_PAIRP (next)) {
      SWiT tmp = SWIT_CDR(next);
      SET_CDR(next, first);
      first = next;
      next = tmp;
    }
    SET_CDR(arg, next);
    return first;
  } else {
    return arg;
  }
}

static SWiT
nlist_insert_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  SWiT lst = arg2;
  if SWIT_FIXNUMP(arg3) {
    int k = SWIT_FIXNUM(arg3);
    register i;
    if (k == 0) {
      return swit_make_pair(arg1, arg2);
    } else {
      for (i = 0; i < k-1; i++) {
	if SWIT_PAIRP(lst) {
	  lst = SWIT_CDR(lst);
	} else {
	  prim_err_msg = index_too_large;
	  exn = exn_range;
	  exn_val = arg3;
	  return NULL;
	}
      }
    }
    if SWIT_PAIRP(lst) {
      SET_CDR(lst, swit_make_pair(arg1, SWIT_CDR(lst)));
      return arg2;
    } else {
      prim_err_msg = not_a_pair;
      exn = exn_type;
      exn_val = arg2;
      return NULL;
    }
  } else if SWIT_PROCP(arg3) {
    SWiT l1 = swit_null_list;
  loop:
    if SWIT_PAIRP(lst) {
      SWiT true;
      SWiT val = SWIT_CAR(lst);
      if ((true = swit_apply (arg3, 1, &val)) != swit_false) {
	if (!true) return NULL;
	if (l1 == swit_null_list) {
	  return swit_make_pair(arg1, lst);
	} else {
	  SET_CDR(l1, swit_make_pair(arg1, lst));
	  return arg2;
	}
      } else {
	l1 = lst;
	lst = SWIT_CDR(lst);
	goto loop;
      }
    } else {
      if SWIT_PAIRP(l1) {
	SET_CDR(l1, swit_make_pair(arg1, swit_null_list));
	return arg2;
      } else {
	return (swit_make_pair(arg1, swit_null_list));
      }
    }
  } else {
    prim_err_msg = wrong_type;
    exn = exn_type;
    exn_val = arg3;
    return NULL;
  }
}

static SWiT
copy_list_prim (SWiT lst)
{
  if (SWIT_NULLP(lst)) {
    return (swit_null_list);
  } else {
    SWiT cdr = SWIT_CDR(lst);
    SWiT last, first, pair;
    first = last = swit_make_pair(SWIT_CAR(lst), cdr);
    SET_CAR(first, (SWiT)(((unsigned int)SWIT_CAR(lst)) | ((unsigned int)(lst->data) & 0x7)));
    lst = cdr;
    while SWIT_PAIRP(lst) {
      cdr = SWIT_CDR(lst);
      pair = swit_make_pair(SWIT_CAR(lst), cdr);
      SET_CAR(pair, (SWiT)(((unsigned int)SWIT_CAR(lst)) | ((unsigned int)(lst->data) & 0x7)));
      SET_CDR (last, pair);
      last = pair;
      lst = cdr;
    }
    return(first);
  }
}

static SWiT
acons_prim (SWiT arg1, SWiT arg2, SWiT arg3)
{
  return (swit_make_pair(swit_make_pair(arg1, arg2), arg3));
}

/*---------------------------------------------------------
 * graphs and marks (for copy, equal? and print)
 *---------------------------------------------------------
 */

#define M(cell) (SWIT_CAR(cell))
#define A(cell) (SWIT_CAR(SWIT_CDR(cell)))
#define D(cell) (SWIT_CDR(SWIT_CDR(cell)))


/* the parameter m is for debugging only */
static SWiT
create_mark (int m)
{
  SWiT obj = swit_alloc_object();
  SWIT_TYPE(obj) = MARK;
  SWIT_FIXNUM(obj) = m;
  return (obj);
}

static int
terminal_p (SWiT x)
{
  return (SWIT_PAIRP(x) ? (SWIT_MARKP(SWIT_CAR(x)) ? 1 : 0) : 1);
}

static SWiT
mkmark (SWiT cell, SWiT m)
{
  SWiT t1 = SWIT_CAR(cell);
  SET_CAR(cell, m);
  SET_CDR(cell, swit_make_pair(t1, SWIT_CDR(cell)));
  return (cell);
}

static SWiT
image (SWiT x)
{
  return (SWIT_PAIRP(x) ? SWIT_CDR(x) : x);
}

void
mark (SWiT expr)
{
  mkmark(expr, EX);
  if (terminal_p(A(expr))) {
    if (terminal_p(D(expr))) {
      SET_CAR(expr, M11);
    } else {
      SET_CAR(expr, M10);
      mark(D(expr));
    }
  } else {
    mark(A(expr));
    if (terminal_p(D(expr))) {
      SET_CAR(expr, M01);
    } else {
      SET_CAR(expr, M00);
      mark(D(expr));
    }
  }
}

SWiT
peel (SWiT oldcel)
{
  SWiT newcel = SWIT_CDR(oldcel);
  SWiT newcar = image(SWIT_CAR(newcel));
  SWiT newcdr = image(SWIT_CDR(newcel));
  SWiT oldcar = SWIT_CAR(newcel);
  SWiT oldcdr = SWIT_CDR(newcel);
  if (M00 == M(oldcel)) {
    peel(oldcdr);
    peel(oldcar);
  } else if (M01 == M(oldcel)) {
    peel(oldcar);
  } else if (M10 == M(oldcel)) {
    peel(oldcdr);
  }
  SET_CAR(oldcel, oldcar);
  SET_CDR(oldcel, oldcdr);
  SET_CAR(newcel, newcar);
  SET_CDR(newcel, newcdr);
  return (newcel);
}

static SWiT
copy_graph_prim (SWiT expr)
{
  if SWIT_PAIRP(expr) {
    mark(expr);
    return (peel(expr));
  } else {
    return (expr);
  }
}


static SWiT
copy_alist_prim (SWiT lst)
{
  if SWIT_NULLP(lst) {
    return (swit_null_list);
  } else {
    SWiT cdr = SWIT_CDR(lst);
    SWiT last, first, pair;
    SWiT car = SWIT_CAR(lst);
    if SWIT_PAIRP(car) {
      car = swit_make_pair(SWIT_CAR(car), SWIT_CDR(car));
    } else {
      prim_err_msg = not_an_alist;
      exn = exn_type;
      exn_val = lst;
      return NULL;
    }
    first = last = swit_make_pair(car, cdr);
    lst = cdr;
    while SWIT_PAIRP(lst) {
      cdr = SWIT_CDR(lst);
      car = SWIT_CAR(lst);
      if (SWIT_PAIRP(car)) {
	car = swit_make_pair(SWIT_CAR(car), SWIT_CDR(car));
      } else {
	prim_err_msg = not_an_alist;
	exn = exn_type;
	exn_val = lst;
	return NULL;
      }
      pair = swit_make_pair(car, cdr);
      SET_CDR (last, pair);
      last = pair;
      lst = cdr;
    }
    return(first);
  }
}

static SWiT
copy_tree_prim (SWiT obj)
{
  if SWIT_PAIRP(obj) {
    return (swit_make_pair (copy_tree_prim(SWIT_CAR(obj)),
			    copy_tree_prim(SWIT_CDR(obj))));
  } else {
    return obj;
  }
}

static SWiT
subst_prim (SWiT new, SWiT old, SWiT tree)
{
  if (swit_equal(old, tree)) {
    return new;
  } else if SWIT_PAIRP(tree) {
    SWiT car = subst_prim (new, old, SWIT_CAR(tree));
    SWiT cdr = subst_prim (new, old, SWIT_CDR(tree));
    if ((car == SWIT_CAR(tree)) && (cdr == SWIT_CDR(tree))) {
      return tree;
    } else {
      return swit_make_pair (car, cdr);
    }
  } else {
    return tree;
  }
}

static SWiT
sublis_prim (SWiT alist, SWiT tree)
{
  SWiT bnd = assoc_prim(tree, alist);
  if (bnd == swit_false) {
    if SWIT_PAIRP(tree) {
      SWiT car = sublis_prim(alist, SWIT_CAR(tree));
      SWiT cdr = sublis_prim(alist, SWIT_CDR(tree));
      if ((car == SWIT_CAR(tree)) && (cdr == SWIT_CDR(tree))) {
	return tree;
      } else {
	return swit_make_pair (car, cdr);
      }
    } else {
      return tree;
    }
  } else {
    if (bnd) {
      return SWIT_CDR(bnd);
    } else {
      prim_err_msg = not_an_alist;
      return NULL;
    }
  }
}

static SWiT
remq_prim (SWiT obj, SWiT list)
{
  if SWIT_PAIRP(list) {
    SWiT car = SWIT_CAR(list);
    SWiT rest = remq_prim(obj, SWIT_CDR(list));
    if (car == obj) {
      return (rest);
    } else {
      return (swit_make_pair(car, rest));
    }
  } else {
    return (list);
  }
}

static SWiT
nremq_prim (SWiT obj, SWiT list)
{
  if (SWIT_PAIRP(list)) {
    SWiT car = SWIT_CAR(list);
    SWiT rest = nremq_prim(obj, SWIT_CDR(list));
    if (car == obj) {
      return (rest);
    } else {
      SET_CDR(list, rest);
      return (list);
    }
  } else {
    return (list);
  }
}

static SWiT
remv_prim (SWiT obj, SWiT list)
{
  if (SWIT_PAIRP(list)) {
    SWiT car = SWIT_CAR(list);
    SWiT rest = remv_prim(obj, SWIT_CDR(list));
    if (swit_eqv(car, obj)) {
      return (rest);
    } else {
      return (swit_make_pair(car, rest));
    }
  } else {
    return (list);
  }
}

static SWiT
nremv_prim (SWiT obj, SWiT list)
{
  if (SWIT_PAIRP(list)) {
    SWiT car = SWIT_CAR(list);
    SWiT rest = nremv_prim(obj, SWIT_CDR(list));
    if (swit_eqv(car, obj)) {
      return (rest);
    } else {
      SET_CDR(list, rest);
      return (list);
    }
  } else {
    return (list);
  }
}

static SWiT
remove_prim (SWiT obj, SWiT list)
{
  if (SWIT_PAIRP(list)) {
    SWiT car = SWIT_CAR(list);
    SWiT rest = remove_prim(obj, SWIT_CDR(list));
    if (swit_equal(car, obj)) {
      return (rest);
    } else {
      return (swit_make_pair(car, rest));
    }
  } else {
    return (list);
  }
}

static SWiT
nremove_prim (SWiT obj, SWiT list)
{
  if (SWIT_PAIRP(list)) {
    SWiT car = SWIT_CAR(list);
    SWiT rest = nremove_prim(obj, SWIT_CDR(list));
    if (swit_equal(car, obj)) {
      return (rest);
    } else {
      SET_CDR(list, rest);
      return (list);
    }
  } else {
    return (list);
  }
}

static SWiT
set_mark_prim (SWiT arg1, SWiT arg2)
{
  SET_CAR(arg1, (SWiT)(((unsigned int)SWIT_CAR(arg1)) & ~0x6));
  SET_CAR(arg1, (SWiT)(((unsigned int)SWIT_CAR(arg1)) | ((((unsigned int)SWIT_FIXNUM(arg2)) << 1) & 0x7)));
  return arg1;
}

static SWiT
get_mark_prim (SWiT arg)
{
  return swit_make_fixnum((((unsigned int)(arg->data)) & 0x6) >> 1);
}



/*==================================================================
 *                       INITIALISATION
 *==================================================================
 */

void
swit_init_pair (void)
{
  read_table['('] = read_pair;
  read_table['['] = read_pair;
  print_table[NIL >> 1] = print_nil;
  print_table[MARK >> 1] = print_mark;
  EX = create_mark(0);
  M00 = create_mark(1);
  M01 = create_mark(2);
  M10 = create_mark(3);
  M11 = create_mark(4);
  swit_make_prim1 ("pair?", pair_p_prim, ANY);
  swit_make_prim2 ("cons", swit_make_pair, ANY, ANY);
  swit_make_prim1 ("car", NULL, 0); /* in eval */
  swit_make_prim1 ("cdr", NULL, 0); /* in eval */
  swit_make_prim2 ("set-car!", set_car_prim, PAIR, ANY);
  swit_make_prim2 ("set-cdr!", set_cdr_prim, PAIR, ANY);
  swit_make_prim1 ("caar", caar_prim, PAIR);
  swit_make_prim1 ("cadr", cadr_prim, PAIR);
  swit_make_prim1 ("cdar", cdar_prim, PAIR);
  swit_make_prim1 ("cddr", cddr_prim, PAIR);
  swit_make_prim1 ("caaar", caaar_prim, PAIR);
  swit_make_prim1 ("caadr", caadr_prim, PAIR);
  swit_make_prim1 ("cadar", cadar_prim, PAIR);
  swit_make_prim1 ("caddr", caddr_prim, PAIR);
  swit_make_prim1 ("cdaar", cdaar_prim, PAIR);
  swit_make_prim1 ("cdadr", cdadr_prim, PAIR);
  swit_make_prim1 ("cddar", cddar_prim, PAIR);
  swit_make_prim1 ("cdddr", cdddr_prim, PAIR);
  swit_make_prim1 ("caaaar", caaaar_prim, PAIR);
  swit_make_prim1 ("caaadr", caaadr_prim, PAIR);
  swit_make_prim1 ("caadar", caadar_prim, PAIR);
  swit_make_prim1 ("caaddr", caaddr_prim, PAIR);
  swit_make_prim1 ("cadaar", cadaar_prim, PAIR);
  swit_make_prim1 ("cadadr", cadadr_prim, PAIR);
  swit_make_prim1 ("caddar", caddar_prim, PAIR);
  swit_make_prim1 ("cadddr", cadddr_prim, PAIR);
  swit_make_prim1 ("cdaaar", cdaaar_prim, PAIR);
  swit_make_prim1 ("cdaadr", cdaadr_prim, PAIR);
  swit_make_prim1 ("cdadar", cdadar_prim, PAIR);
  swit_make_prim1 ("cdaddr", cdaddr_prim, PAIR);
  swit_make_prim1 ("cddaar", cddaar_prim, PAIR);
  swit_make_prim1 ("cddadr", cddadr_prim, PAIR);
  swit_make_prim1 ("cdddar", cdddar_prim, PAIR);
  swit_make_prim1 ("cddddr", cddddr_prim, PAIR);
  swit_make_prim1 ("null?", null_p_prim, ANY);
  swit_make_prim1 ("list?", list_p_prim, ANY);
  swit_make_prim2 ("memq", memq_prim, ANY, LIST);
  swit_make_prim2 ("memv", memv_prim, ANY, LIST);
  swit_make_prim2 ("member", member_prim, ANY, LIST);
  swit_make_prim2 ("assq", assq_prim, ANY, LIST);
  swit_make_prim2 ("assv", assv_prim, ANY, LIST);
  swit_make_prim2 ("assoc", assoc_prim, ANY, LIST);
  swit_make_prim2 ("list-tail", list_tail_prim, PAIR, FIXNUM);
  swit_make_prim2 ("list-ref", list_ref_prim, PAIR, FIXNUM);
  swit_make_prim1 ("reverse", reverse_prim, LIST);
  swit_make_prim1 ("length", length_prim, LIST);
  swit_make_prim0n ("list", list_prim,  ANY, ANY);
  swit_make_prim0n ("append", append_prim,  LIST, ANY);
  /* Extensions */
  swit_make_prim2 ("rassq", rassq_prim, ANY, LIST);
  swit_make_prim2 ("rassv", rassv_prim, ANY, LIST);
  swit_make_prim2 ("rassoc", rassoc_prim, ANY, LIST);
  swit_make_prim0n ("append!", nconc_prim,  LIST, LIST);
  swit_make_prim1 ("reverse!", nreverse_prim, LIST);
  swit_make_prim3 ("list-insert!", nlist_insert_prim, ANY, PAIR, ANY);
  swit_make_prim1 ("copy-list", copy_list_prim, LIST);
  swit_make_prim1 ("copy-alist", copy_alist_prim, LIST);
  swit_make_prim1 ("copy-tree", copy_tree_prim, ANY);
  swit_make_prim1 ("copy-graph", copy_graph_prim, ANY);
  swit_make_prim1 ("last-pair", swit_last_pair, PAIR);
  swit_make_prim2n ("cons*", liststar_prim,  ANY, ANY, ANY, ANY);
  swit_make_prim3 ("acons", acons_prim, ANY, ANY, ANY);
  swit_make_prim2 ("remq", remq_prim, ANY, LIST);
  swit_make_prim2 ("remv", remv_prim, ANY, LIST);
  swit_make_prim2 ("remove", remove_prim, ANY, LIST);
  swit_make_prim2 ("remq!", nremq_prim, ANY, LIST);
  swit_make_prim2 ("remv!", nremv_prim, ANY, LIST);
  swit_make_prim2 ("remove!", nremove_prim, ANY, LIST);
  swit_make_prim3 ("subst", subst_prim, ANY, ANY, LIST);
  swit_make_prim2 ("sublis", sublis_prim, LIST, ANY);
  swit_make_prim1 ("get-mark", get_mark_prim, PAIR);
  swit_make_prim2 ("set-mark!", set_mark_prim, PAIR, FIXNUM);
}
