/* (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include "swit.h"
#include <errno.h>
#include <sys/time.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <memory.h>
#include <fcntl.h>
#include <arpa/inet.h>

extern int h_errno;
extern SWiT swit_make_port (int filedes, io_fun rd, io_fun wr, seek_fun seek);
extern SWiT close_port_prim (SWiT arg);


/*----------------------------------
 *    DEFINITION AND CONSTRUCTION
 *----------------------------------
 */
#define SWIT_SOCKET(obj) ((struct SWiT_Socket *)((obj)->data))
#define SOCKET_PORTNUM(sock) ((sock)->portnum)
#define SOCKET_HOSTNAME(sock) ((sock)->hostname)
#define SOCKET_PORT(sock) ((sock)->port)
#define SOCKET_DESCR(sock) PORT_DESCR(SWIT_PORT(SOCKET_PORT(sock)))

struct SWiT_Socket
{
  unsigned short int portnum;
  SWiT hostname;
  SWiT port;
};
typedef struct SWiT_Socket *SWiT_Socket;

static SWiT
make_socket (unsigned short int portnum, SWiT hostname, int descr)
{
  SWiT obj = swit_alloc_object();
  SWiT_Socket z = SWIT_ALLOC(Socket);
  SOCKET_PORTNUM(z) = portnum;
  SOCKET_HOSTNAME(z) = hostname;
  SOCKET_PORT(z) = swit_make_port(descr, (io_fun)read, (io_fun)write, NULL);
  SWIT_TYPE(obj) = SOCKET;
  SWIT_SOCKET (obj) = z;

  return (obj);
}

/* return 1 if `des' is a socket descriptor, 0 otherwise */
int
issocket(int des)
{
  int val = 0, len;
  return (getsockopt(des, SOL_SOCKET, SO_ERROR, &val, &len) == 0);
}


/*----------------------------------------------------------------
 *    SYNTAX : #<socket <hostname>:<port number> <filedes>>
 *----------------------------------------------------------------
 */
void
swit_print_socket(SWiT sock, int esc, int tcl, SWiT fv, int depth)
{
  SWiT_Socket sh = SWIT_SOCKET(sock);
  str2buff("#<socket ");           /* unreadable */
  print_to_buffer(SOCKET_HOSTNAME(sh), 0, tcl, fv, depth);
  char2buff(':');
  print_to_buffer(swit_make_fixnum((int)SOCKET_PORTNUM(sh)), 0, tcl, fv, depth);
  char2buff(' ');
  print_to_buffer(SOCKET_PORT(sh), 0, tcl, fv, depth);
  char2buff('>');
}


/*-----------------------------------------------------
 *    socket?
 *-----------------------------------------------------
 */
static SWiT
socketp_prim(SWiT sock)
{
  return SWIT_SOCKETP(sock)? swit_true: swit_false;
}


/*-----------------------------------------------------
 *    socket-portnum
 *-----------------------------------------------------
 */
static SWiT
socket_port_number_prim(SWiT sock)
{
  return swit_make_fixnum((int)SOCKET_PORTNUM(SWIT_SOCKET(sock)));
}

/*-----------------------------------------------------
 *    socket-port
 *-----------------------------------------------------
 */
static SWiT
socket_port_prim(SWiT sock)
{
  return SOCKET_PORT(SWIT_SOCKET(sock));
}

/*-----------------------------------------------------
 *    make-server
 *-----------------------------------------------------
 */
static SWiT
make_server_socket_prim(SWiT arg)
{
  struct sockaddr_in sin;
  /* struct servent *service;*/
  int sock, len = sizeof(sin);
  unsigned short int portnum  = (unsigned short int)SWIT_FIXNUM(arg);

  /* some ports are reserved */
  if (getservbyport(htons(portnum), "tcp") != NULL) {
    prim_err_msg = already_reserved;
    exn = exn_io;
    return NULL;
  }
    
  
  /* Create a socket */
  if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  }
  
  /*if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &val, 1) == -1) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
    }*/

  /* Bind a name to the socket */
  memset((char *) &sin, '\0', sizeof(sin));
  sin.sin_family      = AF_INET;
  sin.sin_port 	      = htons(portnum);
  sin.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(sock, (struct sockaddr *) &sin, len) < 0) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  }

  /* Query the socket name
   * (permits to get the true socket number if 0 was given)
   
  if (getsockname(sock, (struct sockaddr *)&sin, &len) < 0) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  } 
  */

  /* Enable the socket to listen for connections (=> server) */
  if (listen(sock, 5) < 0) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  }

  /* Now we can create the socket object */
  return make_socket(ntohs(sin.sin_port),
		     swit_make_constant_string("localhost", 9),
		     sock);
}



/*-----------------------------------------------------
 *    accept
 *-----------------------------------------------------
 */
static SWiT
socket_accept_prim(SWiT sock)
{
  struct sockaddr_in client;
  int s, len;  
  /*  struct hostent *hp; */
  SWiT_Socket s0 = SWIT_SOCKET(sock);
  char *addr, *client_name;
  
  if ((s = accept(SOCKET_DESCR(s0), (struct sockaddr *) &client, &len)) < 0) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  }

  addr = inet_ntoa(client.sin_addr);
  client_name = swit_malloc(strlen(addr));
  strcpy(client_name, addr);
  /*  free(addr);*/

  return make_socket(SOCKET_PORTNUM(s0), swit_make_string(client_name, strlen(client_name)), s);
}


/*-----------------------------------------------------
 *    socket-host
 *-----------------------------------------------------
 */
static SWiT
socket_host_prim(SWiT sock)
{
  return SOCKET_HOSTNAME(SWIT_SOCKET(sock));
}


/*-----------------------------------------------------
 *    make-client
 *-----------------------------------------------------
 */
static SWiT
make_client_socket_prim (SWiT hostname, SWiT port)
{
  struct hostent *hostinfo;
  struct sockaddr_in server;
  int s;
  char *str = STRING_ELS(hostname);
  if (isdigit(*str)) {
    if ((hostinfo = gethostbyaddr(str, strlen(str), AF_INET)) == NULL) {
      prim_err_msg = sys_errlist[h_errno];
      exn = exn_io;
      return NULL;
    }
  } else {
    if ((hostinfo = gethostbyname(str)) == NULL) {
      prim_err_msg = sys_errlist[h_errno];
      exn = exn_io;
      return NULL;
    }
  }

  if ((s = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  }

  server.sin_family = hostinfo->h_addrtype;
  server.sin_port   = htons((unsigned short int)SWIT_FIXNUM(port));
  server.sin_addr   = *(struct in_addr *)hostinfo->h_addr;

  if (connect(s, (struct sockaddr *)&server, sizeof(server)) < 0) {
    printf("2: oops\n");
    prim_err_msg = sys_errlist[errno];
    exn = exn_io;
    return NULL;
  }
  return make_socket(ntohs(server.sin_port), hostname, s);
}


/*-----------------------------------------------------
 *    addr->hostname
 *-----------------------------------------------------
 */
static SWiT
addr_to_name_prim (SWiT arg)
{
  struct hostent *hp;
  struct in_addr iaddr;
  const char *addr = STRING_ELS(arg);

  /* check if the address is valid */
  /* if (inet_aton(addr, &iaddr) == 0) {
    exn = exn_unknown;
    prim_err_msg = wrong_address;
    return NULL;
    }
  */
  if ((hp = gethostbyaddr(addr, STRING_LENGTH(arg), AF_INET)) == NULL) {
    return swit_null_string;
  } else {
    char *name = swit_malloc(strlen(hp->h_name));
    strcpy(name, hp->h_name);
    return swit_make_string(name, strlen(name));
  }
}

/*-----------------------------------------------------
 *    hostname->addr
 *-----------------------------------------------------
 */
static SWiT
name_to_addr_prim (SWiT arg)
{
  char *name = STRING_ELS(arg);
  struct hostent *hostinfo;
  if ((hostinfo = gethostbyname(name)) == NULL) {
    return swit_null_string;
  } else {
    char *addr = inet_ntoa(*(struct in_addr *)hostinfo->h_addr);
    return swit_make_string(addr, strlen(addr));
  }
}

/*-----------------------------------------------------
 *    service->portnum
 *-----------------------------------------------------
 */
static SWiT
name_to_port_prim (SWiT arg)
{
  char *name = STRING_ELS(arg);
  struct servent *s;
  if ((s = getservbyname(name, "tcp")) == NULL) {
    return swit_zero;
  }
  return swit_make_fixnum(ntohs(s->s_port));
}


/*-----------------------------------------------------
 *    portnum->service
 *-----------------------------------------------------
 */
static SWiT
port_to_name_prim (SWiT arg)
{
  int port = SWIT_FIXNUM(arg);
  struct servent *s;
  char *name;
  if ((s = getservbyport(htons(port), "tcp")) == NULL) {
    return swit_null_string;
  }
  name = s->s_name;
  return swit_make_constant_string(name, strlen(name));
}


/*==============================================================
 *                          TYPE
 *-----------------------------------------------------
 */
SWiT swit_socket;

/*-----------------------------------------------------
 *    socket-port?
 *-----------------------------------------------------
 */
static SWiT
socket_port_p_prim (SWiT arg)
{
  if SWIT_PORTP(arg) {
    return (issocket(PORT_DESCR(SWIT_PORT(arg))) ?
	    swit_true : swit_false);
  } else {
    return swit_false;
  }
}


/*===========================
 *    INITIALISATION
 *===========================
 */
void
swit_init_socket(void)
{
  print_table[SOCKET >> 1] = swit_print_socket;
  swit_socket = make_type("socket>");

  swit_make_prim1("socket?", socketp_prim,
		  ANY);
  swit_make_prim1("socket-portnum", socket_port_number_prim,
		  SOCKET);
  swit_make_prim1("socket-port", socket_port_prim,
		  ANY);
  swit_make_prim1("socket-port?", socket_port_p_prim,
		  ANY);
  swit_make_prim1("socket-host", socket_host_prim,
		  SOCKET);
  swit_make_prim1("make-server", make_server_socket_prim,
		   FIXNUM);
  swit_make_prim1("accept", socket_accept_prim,
		  SOCKET);
  swit_make_prim2("make-client", make_client_socket_prim, 
		  STRING_M, FIXNUM);
  swit_make_prim1("service->portnum", name_to_port_prim,
		  STRING_M);
  swit_make_prim1("portnum->service", port_to_name_prim,
		  FIXNUM);
  swit_make_prim1("addr->hostname", addr_to_name_prim,
		  STRING_M);
  swit_make_prim1("hostname->addr", name_to_addr_prim,
		  STRING_M);
}
