/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <tk.h>
#include <errno.h>
#include <math.h>
#include <readline/readline.h>
#include <termios.h>
#include "swit.h"

/* include bitmaps here :
#include "bitmaps/clefa.bmp"
*/

struct termios saved_attributes;
void
reset_attributes (void)
{
  tcsetattr(STDIN_FILENO, TCSANOW, &saved_attributes);
}

extern int string_seek (int descr, int offset, int mode);
extern char *prim_er_msg;
extern int swit_list_length (SWiT arg);
extern void reset_globals(void);

extern SWiT cur_out_port;
#define current_op SWIT_PORT(cur_out_port)
#define print_buffer PORT_BUFFER(current_op)
#define print_cur_idx PORT_INDEX(current_op)
#define out_descr PORT_DESCR(current_op)
#define write_fun PORT_WRITE_FUN(current_op)

extern SWiT cur_in_port;
#define current_ip SWIT_PORT(cur_in_port)
#define read_jbuf PORT_JBUF(current_ip)

extern SWiT swit_load (char *file);
extern SWiT string_to_port_prim(SWiT arg);
extern SWiT make_fv(SWiT_Env e);
extern void swit_load_init (char *filename);
extern int interrupt;
extern char *tok_text;

static Tcl_Interp *tks_interp;
static SWiT tks_port;
#define tks_io SWIT_PORT(tks_port)
#define	tks_in_buf PORT_BUFFER(tks_io)
#define BUFFER_SIZE 0xffff
static char swittk_buffer[BUFFER_SIZE];

int tk;

static int
tcl_print_to_buffer (SWiT obj)
{
  char *err_msg;
  int len, idx;
  SWiT old_op = cur_out_port;
  cur_out_port = tks_port;
  print_cur_idx = 0;
  if (string_seek(out_descr, 0, SEEK_SET) == -1) {
    goto erreur;
  }
  /* Must omit outside parens for a list or a vector
   * (and handle these special cases for 'toplevel' display)
   */
  if SWIT_PAIRP(obj) {
    do {
      print_to_buffer(SWIT_CAR(obj), 0, 1, swit_null_list, 0);
      obj = SWIT_CDR(obj);
      if (obj != swit_null_list) {
	space2buff ();
      }
    } while SWIT_PAIRP(obj);
    if (obj != swit_null_list) {
      str2buff(". ");
      print_to_buffer(obj, 0, 1, swit_null_list, 0);
    }
  } else if SWIT_VECTORP(obj) {
    int size = VECTOR_LENGTH(obj);
    if (size != 0) {
      SWiT *els = VECTOR_ELS(obj);
      int i;
      print_to_buffer(els[0], 0, 1, swit_null_list, 0);
      for (i = 1; i < size; ++i) {
	space2buff ();
	print_to_buffer(els[i], 0, 1, swit_null_list, 0);
      }
    }
  } else {
    print_to_buffer(obj, 0, 1, swit_null_list, 0);
  }
  len = print_cur_idx;
  if ((*write_fun)(out_descr, print_buffer, len) != len) {
    goto erreur;
  }
  if ((idx = string_seek(out_descr, 0, SEEK_CUR)) == -1) {
    goto erreur;
  }

  swittk_buffer[idx] = '\0';
  cur_out_port = old_op;
  return 1;

 erreur:
  err_msg = "cannot output result";
  fprintf(stderr, "[TCL ERROR] -- ");
  fprintf(stderr, err_msg);
  fprintf(stderr, "\n");
  return 0;
}

static void
Tk_DoNothing (ClientData clientData, int mask)
{
}

static int
Tk_rl_DoOneEvent (void)
{
  Tcl_DoOneEvent(0);
  return 0;
}

static char *
strlower (char *str)
{
  register char *s = str;
  for (;(*s=tolower(*s));s++) {
    continue;
  }
  return str;
}


int
Swit_CallCmd(ClientData env, Tcl_Interp *interp, int argc, char *argv[])
{
  SWiT old_ip = cur_in_port;
  int nargs = argc-2;

  if (argc < 2) {
    char *err_msg = swit_call_wrong_number;
    Tcl_SetResult(interp, err_msg, TCL_STATIC);
    fprintf(stderr, "[TCL ERROR] -- ");
    fprintf(stderr, err_msg, tok_text);
    fprintf(stderr, "\n");
    return TCL_ERROR;
  } else {
    char *fun_name = argv[1];
    SWiT fun_sym = swit_make_symbol(strlower(fun_name), strlen(fun_name), swit_unbound);
    SWiT fun = SYMBOL_VAL(SWIT_SYM(fun_sym));
    if (!SWIT_PROCP(fun)) {
      char *err_msg = swit_call_isnot_fun;
      Tcl_SetResult(interp, err_msg, TCL_STATIC);
      fprintf(stderr, "[TCL ERROR] -- ");
      fprintf(stderr, err_msg, tok_text);
      fprintf(stderr, tk_in);
      fprintf(stderr, "(%s ...)\n", fun_name);
      return TCL_ERROR;
    }
    tk = 1;
    cur_in_port = tks_port;
    if (setjmp(*read_jbuf)) {
      Tcl_SetResult(interp, prim_err_msg, TCL_STATIC);
      cur_in_port = old_ip;
      fprintf(stderr, "[TCL ERROR] -- (swit-call) ");
      fprintf(stderr, prim_err_msg, tok_text);
      fprintf(stderr, tk_in);
      fprintf(stderr, "(%s ...)\n", fun_name);
      tk = 0;
      return TCL_ERROR;
    }
    if (nargs != 0) {
      SWiT vals[nargs];
      register int i;
      argv +=2 ;
      for (i = 0; i < nargs; i++) {
	SWiT obj;
	char *str = argv[i];
	int len = strlen(str);
	/* ``read'' argv[i] into tks_port */
	register j;
	for(j = 0; j < len; j++) {
	  tks_in_buf[j] = *str++;
	}
	tks_in_buf[j] = ' '; /* a separator */
	PORT_INDEX(tks_io) = 0;
	PORT_EOB(tks_io) = len+1;
	obj = rd_datum(get_tok());
	if SWIT_SYMBOLP(obj) {
	  vals[i] = SYMBOL_VAL(SWIT_SYM(obj));
	} else {
	  vals[i] = obj;  /* a literal : not evaluated ! */
	}
      }
      cur_in_port = old_ip;
      if (!swit_apply(fun, nargs, vals)) { /* result ignored */
	Tcl_SetResult(interp, prim_err_msg , TCL_STATIC);
	fprintf(stderr, "[TCL ERROR] -- (swit-call) ");
	fprintf(stderr, prim_err_msg, tok_text);
	fprintf(stderr, tk_in);
	fprintf(stderr, "(%s %s ...)\n", fun_name, *argv);
	tk = 0;
	return TCL_ERROR;
      }
    } else {
      if (!swit_apply(fun, 0, &swit_null_list)) {  /* result ignored */
	Tcl_SetResult(interp, prim_err_msg, TCL_STATIC);
	cur_in_port = old_ip;
	fprintf(stderr, "[TCL ERROR] -- (swit-call) ");
	fprintf(stderr, prim_err_msg, tok_text);
	fprintf(stderr, tk_in);
	fprintf(stderr, "(%s)\n", fun_name);
	tk = 0;
	return TCL_ERROR;
      }
    }
    /* BEWARE : no result (for sake of efficiency)*/
    Tcl_SetResult(interp, "", TCL_STATIC);
    cur_in_port = old_ip;
    tk = 0;
    return TCL_OK;
  }
}


int
Swit_ValCmd(ClientData env, Tcl_Interp *interp, int argc, char *argv[])
{
  if (argc != 2) {
    char *err_msg = swit_val_wrong_number;
    Tcl_SetResult(interp, err_msg, TCL_STATIC);
    fprintf(stderr, "[TCL ERROR] -- ");
    fprintf(stderr, err_msg);
    fprintf(stderr, "\n");
    return TCL_ERROR;
  } else {
    SWiT obj = swit_make_symbol(argv[1], strlen(argv[1]),swit_unbound);
    SWiT val = SYMBOL_VAL(SWIT_SYM(obj));
    if (val == swit_unbound) {
      char *err_msg = swit_val_not_bound;
      Tcl_SetResult(interp, err_msg, TCL_STATIC);
      fprintf(stderr, "[TCL ERROR] -- ");
      fprintf(stderr, err_msg);
      fprintf(stderr, " `");
      fprintf(stderr, SYMBOL_STRING(SWIT_SYM(obj)));
      fprintf(stderr, "'\n");
      return TCL_ERROR;
    } else {
      if (tcl_print_to_buffer(val)) {
	Tcl_SetResult(interp, swittk_buffer, TCL_STATIC);
	return TCL_OK;
      }
      return TCL_ERROR;
    }
  }
}

extern SWiT_Ccont done_cont;

int
Swit_ExprCmd(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
  if (argc < 2) {
    char *err_msg = swit_expr_wrong_number;
    Tcl_SetResult(interp, err_msg , TCL_STATIC);
    fprintf(stderr, "[TCL ERROR] -- ");
    fprintf(stderr, err_msg);
    fprintf(stderr, "\n");
    return TCL_ERROR;
  } else {
    SWiT_Expr obj;
    SWiT ret = swit_null_list;
    SWiT_Port old_ip = current_ip;
    int len, i;
    char *str = argv[1];
    tk = 1;
    /* ``read'' argv[1] into tks_port */
    len = strlen(str);
    for(i = 0; i < len; i++) {
      tks_in_buf[i] = *str++;
    }
    tks_in_buf[i] = ' ';   /* a separator */
    PORT_INDEX(tks_io) = 0;
    PORT_EOB(tks_io) = len+1;
    if (argc == 2) {
      current_ip = tks_io;
      obj = read_program(get_tok());
      current_ip = old_ip;
      ret = swit_eval(obj, NULL, done_cont);
      if (!ret) {
	fprintf(stderr, "[TCL ERROR] -- (swit-expr) ");
	fprintf(stderr, prim_err_msg, tok_text);
	fprintf(stderr, tk_in);
	fprintf(stderr, "%s\n", argv[1]);
	Tcl_SetResult(interp, prim_err_msg, TCL_STATIC);
	current_ip = old_ip;
	return TCL_ERROR;
      }
    } else if (argc == 3) /* argv[2] must be a symbol bound to an env */ {
      SWiT env_sym = swit_make_symbol(argv[2], strlen(argv[2]), swit_unbound);
      SWiT env = SYMBOL_VAL(SWIT_SYM(env_sym));
      if (!SWIT_ENVP(env)) {
	char *err_msg = swit_expr_no_env;
	Tcl_SetResult(interp, err_msg, TCL_STATIC);
	fprintf(stderr, "[TCL ERROR] -- ");
	fprintf(stderr, err_msg, SYMBOL_STRING(SWIT_SYM(env_sym)));
	fprintf(stderr, "\n");
	tk = 0;
	return TCL_ERROR;
      }
      current_ip = tks_io;
      obj = read_exp(get_tok(), env);
      current_ip = old_ip;
      ret = swit_eval(obj, SWIT_ENV(env), done_cont);
      if (!ret) {
	fprintf(stderr, "[TCL ERROR] -- (swit-expr) ");
	fprintf(stderr, prim_err_msg, tok_text);
	fprintf(stderr, tk_in);
	fprintf(stderr, "%s\n", argv[1]);
	Tcl_SetResult(interp, prim_err_msg, TCL_STATIC);
	current_ip = old_ip;
	return TCL_ERROR;
      }
    } else /* error */ {
      char *err_msg = swit_expr_wrong_number;
      Tcl_SetResult(interp, err_msg, TCL_STATIC);
      fprintf(stderr, "[TCL ERROR] -- ");
      fprintf(stderr, err_msg);
      fprintf(stderr, "\n");
      tk = 0;
      return TCL_ERROR;
    }
    if (tcl_print_to_buffer(ret)) {
      Tcl_SetResult(interp, swittk_buffer, TCL_STATIC);
      tk = 0;
      return TCL_OK;
    }
    tk = 0;
    return TCL_ERROR;
  }
}

/*
 * tk-eval is a scheme primitive, its argument is a tcl/tk script.
 *
 * Type : string -> string
 * Return : the result stored in tks_interp->result
 */
static SWiT
tk_eval (SWiT arg)
{
  char *cmd = STRING_ELS(arg);
  interrupt = 0;
  Tcl_CreateFileHandler(0, 0, Tk_DoNothing, (ClientData) 0);
  if (Tcl_Eval(tks_interp, cmd) == TCL_OK) {
    int len = strlen(tks_interp->result);
    char *res = (char *)swit_malloc_atomic(len+1);
    SWiT ret;
    memcpy(res, tks_interp->result, len);
    res[len] = '\0';
    ret = swit_make_string(res, len);
    Tcl_CreateFileHandler(0, TCL_READABLE, Tk_DoNothing, (ClientData) 0);
    Tcl_ResetResult(tks_interp);
    interrupt = 1;
    return ret;
  }
  /* error */
  /*  Tcl_VarEval(tks_interp,"tkerror {", tks_interp->result, "}", (char *)NULL);*/

  Tcl_CreateFileHandler(0, TCL_READABLE, Tk_DoNothing, (ClientData) 0);
  reset_globals();

  /*  return swit_null_string;*/
  prim_err_msg = tks_interp->result;
  exn = exn_unknown;
  exn_val = arg;
  interrupt = 1;
  return NULL;
}

/*
 * tk-cmd is a scheme primitive, its argument is a tcl/tk script.
 * same as tk-eval but `no' result returned
 * Type : string -> void
 * Return : #t or #f
 */
static SWiT
tk_cmd (SWiT arg)
{
  char *cmd = STRING_ELS(arg);
  int res;
  interrupt = 0;
  Tcl_CreateFileHandler(0, 0, Tk_DoNothing, (ClientData) 0);
  res = Tcl_Eval(tks_interp, cmd);
  if (res == TCL_OK) {
    Tcl_CreateFileHandler(0, TCL_READABLE, Tk_DoNothing, (ClientData) 0);
    Tcl_ResetResult(tks_interp);
    interrupt = 1;
    return swit_true;
  }
  /* error */
  /*  Tcl_VarEval(tks_interp,"tkerror {", tks_interp->result, "}", (char *)NULL);*/

  Tcl_CreateFileHandler(0, TCL_READABLE, Tk_DoNothing, (ClientData) 0);
  reset_globals();

  /* return false in an error occured */
  /*  return swit_false;*/
  prim_err_msg = tks_interp->result;
  exn = exn_unknown;
  exn_val = arg;
  interrupt = 1;
  return NULL;
}

/*-------------------------------------------------------------
 * the argument can be a filename relative to SWIT_LIBRARY_PATH
 * or to the current directory if this environment variable is
 * not defined or filename begins with a '.' or a '..', or an
 * absolute filename if it starts with a '/'.
 *-------------------------------------------------------------
 */
extern SWiT swit_load_path;
static SWiT
tk_source (SWiT arg)
{
  char buffer[256] = "source ";
  char *cmd = buffer;
  char *str = STRING_ELS(arg);
  int i,len=7;
  for (i = 7; i < 256; i++) buffer[i] = 0;
  if ((*str != '/') && (*str != '.')) {
    SWiT dirname = SYMBOL_VAL(SWIT_SYM(swit_load_path));
    int l = STRING_LENGTH(dirname);
    if (l) memcpy(cmd+len, STRING_ELS(dirname), l);
    len += l;
  }
  memcpy(cmd+len, str, STRING_LENGTH(arg));
  /* check */
  fprintf(stderr,"; %s\n", buffer);
  if (Tcl_Eval(tks_interp, buffer) == TCL_OK) {
    return swit_true;
  }
  /* Call tkerror */
  Tcl_VarEval(tks_interp,"tkerror {", tks_interp->result, "}", (char *)NULL);
  /* return false in an error occured */
  return swit_false;
}


static int orig_percent, orig_pitch, orig_duration;

SWiT
swit_ring_bell (void)
{
  Tk_Window tkwin = Tk_MainWindow(tks_interp);
  Display *dpy = Tk_Display(tkwin);
  XBell(dpy, 100);
  return swit_true;
}


/*--------------------------------------------------------------------*
 |                            GetBell                                 |
 *--------------------------------------------------------------------*/
void GetBell(Display *dpy)
{
  XKeyboardState stateValues;
  XGetKeyboardControl(dpy, &stateValues);
  orig_percent  = stateValues.bell_percent;
  orig_pitch    = stateValues.bell_pitch;
  orig_duration = stateValues.bell_duration;
}

SWiT
swit_init_bell (void)
{
  Tk_Window tkwin = Tk_MainWindow(tks_interp);
  Display *dpy = Tk_Display(tkwin);
  GetBell(dpy);
  return swit_make_pair(swit_make_fixnum(orig_percent),
	     swit_make_pair(swit_make_fixnum(orig_pitch),
		   swit_make_pair(swit_make_fixnum(orig_duration),
			swit_null_list)));
}

/*--------------------------------------------------------------------*
 |                             SetBell                                |
 *--------------------------------------------------------------------*/
void SetBell(Display *dpy, int pitch, int duration)
{
  XKeyboardControl controlValues;
  unsigned long valueMask = KBBellPercent | KBBellPitch | KBBellDuration;
  controlValues.bell_percent  = orig_percent;
  controlValues.bell_pitch    = pitch;
  controlValues.bell_duration = duration;
  XChangeKeyboardControl(dpy, valueMask, &controlValues);
}

SWiT
swit_set_bell (SWiT arg)
{
  Tk_Window tkwin = Tk_MainWindow(tks_interp);
  Display *dpy = Tk_Display(tkwin);
  SetBell(dpy, SWIT_FIXNUM(arg), 100);
  return swit_true;
}

#define BASE_F   (double)246.9413

/*--------------------------------------------------------------------*
 |                            Pitch                                   |
 *--------------------------------------------------------------------*/
int Pitch (int note)
{
  double x, m, n, f;

  /* notes are calculated from the base frequency of BASE_F. */
  /* This is the first note on the keyboard.                 */
  /* The frequency of a note = 2^(index / 12).               */

  x = (double)2.0;
  m = (double)note;
  n = (double)12.0;

  f = BASE_F * pow(x, (m/n));

  return((int)f);
}

/*--------------------------------------------------------------------*
 |                             PlayNote                               |
 *--------------------------------------------------------------------*/
void PlayNote (Display *dsp, int note, int duration)
{
  SetBell(dsp, Pitch(note), duration);
  XBell(dsp, 100);
  SetBell(dsp, orig_pitch, orig_duration);
}

SWiT
swit_play_note (SWiT note, SWiT duration)
{
  Tk_Window tkwin = Tk_MainWindow(tks_interp);
  Display *dpy = Tk_Display(tkwin);
  interrupt = 0;
  PlayNote (dpy, SWIT_FIXNUM(note), SWIT_FIXNUM(duration));
  interrupt = 1;
  return swit_true;
}


extern void swit_banner(void);
extern void swit_load_objs(void);

int swit_init_tk(Tcl_Interp *interp)
{
  char *argv0, *appl_name;
  SWiT string_buffer = swit_make_string(swittk_buffer, BUFFER_SIZE);

  /* make interp avaliable globally */
  tks_interp = interp;

  /* Initialize scheme and add the 'tk-xxx' primitives */
  swit_init();

  swit_make_prim1 ("tk-eval", tk_eval,
		   STRING_M);
  swit_make_prim1 ("tk-cmd", tk_cmd,
		   STRING_M);
  swit_make_prim1 ("tk-source", tk_source,
		   STRING_M);
  swit_make_prim0 ("ring-bell", swit_ring_bell);
  swit_make_prim0 ("init-bell", swit_init_bell);
  swit_make_prim1 ("set-bell!", swit_set_bell,
		   FIXNUM);
  swit_make_prim2 ("play-note", swit_play_note,
		   FIXNUM, FIXNUM);

  /* Open the string port to read/write scheme expressions */
  tks_port = string_to_port_prim(string_buffer);

  /* Commands to call scheme from tcl */
  Tcl_CreateCommand(tks_interp, "swit-val", Swit_ValCmd, NULL, NULL);
  Tcl_CreateCommand(tks_interp, "swit-call", Swit_CallCmd, NULL, NULL);
  Tcl_CreateCommand(tks_interp, "swit-expr", Swit_ExprCmd, NULL, NULL);

  Tcl_CreateFileHandler(0, TCL_READABLE, Tk_DoNothing, (ClientData) 0);

  /* load *.obj */
  swit_load_objs();

  /* If a (scheme) file was given as first argument : load it */
  argv0 = Tcl_GetVar(interp,"argv0", TCL_GLOBAL_ONLY);
  appl_name = strrchr(argv0, '/');
  if (appl_name != NULL) {
    appl_name++;
  } else {
    appl_name = argv0;
  }
  if (strcmp(appl_name, "swit") != 0)  {
    swit_load(argv0);
  }

  /*
   * Some bimaps could be defined here :
   Tk_DefineBitmap(interp, Tk_GetUid("clefa"), clefa_bits, clefa_width, clefa_height);
   */

  /* unmap main window (can be remapped by a call : "wm deiconify ." )*/
  Tcl_Eval(tks_interp, "wm withdraw .");

  /* hook for readline */
  rl_event_hook = Tk_rl_DoOneEvent;

  /* enable interrupts */
  interrupt = 1;

  /* errors */
  tk = 0;

  /* save terminal attributes and restore them at exit*/
  tcgetattr(STDIN_FILENO, &saved_attributes);
  on_exit(reset_attributes, 1);

  /* Finally, call the swit read-eval-print loop : */
  swit_repl();

  exit(1);
}
