/*
Copyright (C) 1996,1997,1998,1999 Didier VIDAL

This file is part of Scheme With Types, herein refered as SWiT.

SWiT is free software; you can redistribute it and/or modify it under
the terms of the GNU LIBRARY GENERAL PUBLIC LICENSE as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

SWiT is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.

You should have received a copy of the GNU Library General Public License
along with SWiT software; see the file COPYING.
If not, write to the Free Software
Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "swit.h"
#include <math.h>
#include <sys/socket.h>
#include <sys/times.h>
#include <sys/time.h>

#define current_ip SWIT_PORT(cur_in_port)
#define lineno PORT_LINENO(current_ip)

/* import */
extern int issocket(int des);
extern SWiT make_fv(SWiT_Env e);
extern void swit_print_code (SWiT_Expr obj, SWiT fv);
extern void swit_printl (SWiT obj, SWiT fv);
extern int read_paren;
extern int prompt;
extern char *filename;
extern SWiT default_print_depth;
extern SWiT default_print_span;
extern SWiT swit_print_depth;
extern SWiT swit_print_span;
extern SWiT swit_prompt;
extern SWiT swit_break_prompt;
extern SWiT swit_load_path;
extern SWiT swit_type_flag;
extern SWiT cur_in_port;
extern SWiT cur_out_port;
extern SWiT cur_err_port;
extern int step_start;
extern int transcript_descr;
extern int raw_mode;
extern int tk;
extern SWiT_Exn exn_stack;

static sigset_t mask;
volatile int interrupt;
volatile int error_code;

/* exported */
jmp_buf *error_Ptr;
jmp_buf top_level_buf;
int print_span;
int print_depth;


/* A simplfied version of the preceding procedure
 *  (when the error in certainly not recoverable)
 */
void
swit_signal_error (SWiT err, char *msg, ...)
{
  error_code = 0;
  if (error_Ptr == &top_level_buf) {
    int filedes = PORT_DESCR(SWIT_PORT(cur_in_port));
    int issock = issocket(filedes); 
    va_list args;
    if (!(isatty (filedes)) && !issock && (filedes != -1)) {
      fprintf(stderr, error_at_line, filename, lineno);
    } else {
      fprintf(stderr, "[ERROR] -- ");
    }
    if (msg) {
      va_start (args, msg);
      vfprintf (stderr, msg, args);
      va_end (args);
    } else {
      fprintf(stderr, exception);
      fprintf(stderr, uncaught_exception, (char *)STRING_ELS(exn));
    }
    if (transcript_descr != -1) {
      write(transcript_descr, exception_newline, 17);
    }
    /*    fprintf (stderr, "\n");*/
    if (err && (err != swit_else)) {
      if (msg) fprintf(stderr, in);
      swit_printl(err, swit_null_list);
      fprintf(stderr,"\n");
    }
  }
  longjmp(*error_Ptr, 1);
}

static void handle_sigfpe(int sig)
{
  signal(SIGFPE, handle_sigfpe);
  error_code = 1;
  exn = exn_math;
  longjmp(*error_Ptr, 1);
}

static void handle_sigill(int sig)
{
  signal(SIGILL, handle_sigill);
  error_code = 2;
  longjmp(*error_Ptr, 1);
}

static void handle_sigbus(int sig)
{
  signal(SIGBUS, handle_sigbus);
  error_code = 3;
  longjmp(*error_Ptr, 1);
}

static void handle_sigsegv(int sig)
{
  signal(SIGSEGV, handle_sigsegv);
  error_code = 4;
  longjmp(*error_Ptr, 1);
}


static void
handle_sigint(int sig)
{
  signal(SIGINT, handle_sigint);
  sigprocmask(SIG_UNBLOCK, &mask, NULL);
  if (interrupt) {
    error_code = 5;
    longjmp(*error_Ptr, 1);
  }
}


void
reset_globals (void)
{
  cur_in_port = swit_stdin_port;
  cur_out_port = swit_stdout_port;
  cur_err_port = swit_stderr_port;
  SYMBOL_VAL(SWIT_SYM(swit_load_path)) = swit_null_string;
  swit_prompt = swit_make_symbol("*prompt*", 8, swit_unbound);
  SYMBOL_VAL(SWIT_SYM(swit_print_span)) = default_print_span;
  SYMBOL_VAL(SWIT_SYM(swit_print_depth)) = default_print_depth;
  error_Ptr = &top_level_buf;
  tk = 0;
}

static void
init_vars (void)
{
  /* (re)initialize some variables */
  PORT_INDEX(SWIT_PORT(cur_out_port)) = 0;
  PORT_INDEX(SWIT_PORT(cur_in_port)) = 0;
  PORT_INDEX(SWIT_PORT(cur_err_port)) = 0;
  PORT_EOB(SWIT_PORT(cur_out_port)) = 0;
  PORT_EOB(SWIT_PORT(cur_in_port)) = 0;
  PORT_EOB(SWIT_PORT(cur_err_port)) = 0;
  step_start = 1;
  read_paren = 0;
  loading = 0;
  prompt = 1;
  filename = "";
  prim_err_msg = "";
  interrupt = 1;
  raw_mode = 0;
  exn_stack = NULL;
  exn = exn_unknown;
  error_code = 0;
  tk = 0;
  print_depth = SWIT_FIXNUM(SYMBOL_VAL(SWIT_SYM(swit_print_depth)));
  print_span = SWIT_FIXNUM(SYMBOL_VAL(SWIT_SYM(swit_print_span)));
}

void
swit_init_top (void)
{
  signal(SIGINT, handle_sigint);
  signal(SIGTERM, handle_sigint);
  signal(SIGFPE, handle_sigfpe);
  signal(SIGBUS, handle_sigbus);
  signal(SIGILL, handle_sigill);
  signal(SIGSEGV, handle_sigsegv);
  sigemptyset(&mask);
  sigaddset(&mask, SIGINT);
  reset_globals();
  init_vars();
}

extern SWiT_Expr compile_program (SWiT datum);
extern SWiT_Ccont done_cont;

/*
 * Read-Eval-Print loop
 */
#define current_ip SWIT_PORT(cur_in_port)
#define read_jbuf PORT_JBUF(current_ip)
void
swit_repl (void)
{
  SWiT type = swit_make_symbol(":type", 5, swit_unbound);
  swit_banner();
  while (1) {
    SWiT_Expr tye, expr;
    SWiT val;
    SWiT datum;
    clock_t start, end;
    int flag;
    double time;

    /* interrupts */
    if (setjmp(*error_Ptr)) {
      switch (error_code)
	{
	case 1: fprintf(stderr, arith_error);break;
	case 2: fprintf(stderr, stack_overflow);break;
	case 3: fprintf(stderr, bus_error);break;
	case 4: fprintf(stderr, seg_fault);break;
	case 5: fprintf(stderr, interrupt_received);break;
	default: fprintf(stderr, ret_to_top);
	}
      reset_globals();
    }

    flag = 0;
    init_vars();

    /* read */
    if (setjmp(*read_jbuf)) {
      swit_signal_error(NULL, prim_err_msg, tok_text);
    }
    datum = rd_datum(get_tok());

    /* type (if asked to) */
    if ((SYMBOL_VAL(SWIT_SYM(type)) != type) 
	&& (SYMBOL_VAL(SWIT_SYM(swit_type_flag)) == swit_true)) {
      SWiT_Expr expr = swit_make_quote(datum);
      SWiT expr_type = swit_apply_proc(":type", 1, &expr);
      if (expr_type) {
	tye = swit_make_quote(expr_type);
	flag = 1;
      } else {
	fprintf(stderr, type_impl_err);
	flag = 0;
      }
    }

    /* compile */
    if ((expr = compile_program (datum)) == NULL) {
      swit_signal_error(datum, comp_impl_err);
    }

    /* eval */
    start = clock();
    val = swit_eval(expr, NULL, done_cont);
    end = clock();
    
    /* set *time* variable */
# ifdef LINUX
    /* mal foutu sur Linux : clock() retourne des diximes de ms ... */ 
    time = ((double) (end-start)) * 10;
# else /* sun */
    /* sur sun clock() retourne des micro-secondes */
    time = ((double) (end-start)) / 1000;
# endif
    /* print time if `verbose' was asked */
    if (SYMBOL_VAL(SWIT_SYM(swit_verbose)) == swit_true) {
      fprintf(stdout, temps, time);
    }
    SYMBOL_VAL(SWIT_SYM(swit_time)) = swit_make_inexact(time);

    /* print result */
    fflush(stdout);
    if (val) {
      swit_printl(val, swit_null_list);
      
      /* print type */
      if (flag) {
	swit_apply_proc(":pptype", 1, &tye);
      }
    
      /* newline */
      if (transcript_descr != -1) {
	write(transcript_descr, "\n", 1);
      }
      printf ("\n");
    } else {
      fprintf(stderr, ret_to_top);
    }
  }
}
